# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Example for generating density plots."""

import pylab as plt
import numpy as np

from neurom import get as get_feat
from neurom.view import (common, view)
from neurom.core.types import NeuriteType


def extract_density(population, plane='xy', bins=100, neurite_type=NeuriteType.basal_dendrite):
    """Extracts the 2d histogram of the center
       coordinates of segments in the selected plane.
    """
    segment_midpoints = get_feat('segment_midpoints', population, neurite_type=neurite_type)
    horiz = segment_midpoints[:, 'xyz'.index(plane[0])]
    vert = segment_midpoints[:, 'xyz'.index(plane[1])]
    return np.histogram2d(np.array(horiz), np.array(vert), bins=(bins, bins))


def plot_density(population,  # pylint: disable=too-many-arguments, too-many-locals
                 bins=100, new_fig=True, subplot=111, levels=None, plane='xy',
                 colorlabel='Nodes per unit area', labelfontsize=16,
                 color_map='Reds', no_colorbar=False, threshold=0.01,
                 neurite_type=NeuriteType.basal_dendrite, **kwargs):
    """Plots the 2d histogram of the center
       coordinates of segments in the selected plane.
    """
    fig, ax = common.get_figure(new_fig=new_fig, subplot=subplot)

    H1, xedges1, yedges1 = extract_density(population, plane=plane, bins=bins,
                                           neurite_type=neurite_type)

    mask = H1 < threshold  # mask = H1==0
    H2 = np.ma.masked_array(H1, mask)

    getattr(plt.cm, color_map).set_bad(color='white', alpha=None)

    plots = ax.contourf((xedges1[:-1] + xedges1[1:]) / 2,
                        (yedges1[:-1] + yedges1[1:]) / 2,
                        np.transpose(H2), # / np.max(H2),
                        cmap=getattr(plt.cm, color_map), levels=levels)

    if not no_colorbar:
        cbar = plt.colorbar(plots)
        cbar.ax.set_ylabel(colorlabel, fontsize=labelfontsize)

    kwargs['title'] = kwargs.get('title', '')
    kwargs['xlabel'] = kwargs.get('xlabel', plane[0])
    kwargs['ylabel'] = kwargs.get('ylabel', plane[1])

    return common.plot_style(fig=fig, ax=ax, **kwargs)


def plot_neuron_on_density(population, # pylint: disable=too-many-arguments
                           bins=100, new_fig=True, subplot=111, levels=None, plane='xy',
                           colorlabel='Nodes per unit area', labelfontsize=16,
                           color_map='Reds', no_colorbar=False, threshold=0.01,
                           neurite_type=NeuriteType.basal_dendrite, **kwargs):
    """Plots the 2d histogram of the center
       coordinates of segments in the selected plane
       and superimposes the view of the first neurite of the collection.
    """
    _, ax = common.get_figure(new_fig=new_fig)

    view.plot_tree(ax, population.neurites[0])

    return plot_density(population, plane=plane, bins=bins, new_fig=False, subplot=subplot,
                        colorlabel=colorlabel, labelfontsize=labelfontsize, levels=levels,
                        color_map=color_map, no_colorbar=no_colorbar, threshold=threshold,
                        neurite_type=neurite_type, **kwargs)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Python module of NeuroM to check neuronal trees."""

import numpy as np
from neurom.core.dataformat import COLS
from neurom import morphmath as mm
from neurom.morphmath import principal_direction_extent


def is_flat(neurite, tol, method='tolerance'):
    """Check if neurite is flat using the given method.

    Args:
        neurite(Neurite): neurite to operate on
        tol(float): tolerance
        method(string): the method of flatness estimation:
            'tolerance' returns true if any extent of the tree is smaller
            than the given tolerance
            'ratio' returns true if the ratio of the smallest directions
            is smaller than tol. e.g. [1,2,3] -> 1/2 < tol

    Returns:
        True if neurite is flat
    """
    ext = principal_direction_extent(neurite.points[:, COLS.XYZ])

    assert method in ('tolerance', 'ratio'), "Method must be one of 'tolerance', 'ratio'"
    if method == 'ratio':
        sorted_ext = np.sort(ext)
        return sorted_ext[0] / sorted_ext[1] < float(tol)
    return any(ext < float(tol))


def is_back_tracking(neurite):
    """Check if a neurite process backtracks to a previous node.

    Back-tracking takes place
    when a daughter of a branching process goes back and either overlaps with a previous point, or
    lies inside the cylindrical volume of the latter.

    Args:
        neurite(Neurite): neurite to operate on

    Returns:
        True Under the following scenaria:
            1. A segment endpoint falls back and overlaps with a previous segment's point
            2. The geometry of a segment overlaps with a previous one in the section
    """
    def pair(segs):
        """Pairs the input list into triplets."""
        return zip(segs, segs[1:])

    def coords(node):
        """Returns the first three values of the tree that correspond to the x, y, z coordinates."""
        return node[COLS.XYZ]

    def max_radius(seg):
        """Returns maximum radius from the two segment endpoints."""
        return max(seg[0][COLS.R], seg[1][COLS.R])

    def is_not_zero_seg(seg):
        """Returns True if segment has zero length."""
        return not np.allclose(coords(seg[0]), coords(seg[1]))

    def is_in_the_same_verse(seg1, seg2):
        """Checks if the vectors face the same direction.

        This is true if their dot product is greater than zero.
        """
        v1 = coords(seg2[1]) - coords(seg2[0])
        v2 = coords(seg1[1]) - coords(seg1[0])
        return np.dot(v1, v2) >= 0

    def is_seg2_within_seg1_radius(dist, seg1, seg2):
        """Checks if both segment are within the sum of their radii from one another."""
        return dist <= max_radius(seg1) + max_radius(seg2)

    def is_seg1_overlapping_with_seg2(seg1, seg2):
        """Checks if a segment is in proximity of another one upstream."""
        # get the coordinates of seg2 (from the origin)
        s1 = coords(seg2[0])
        s2 = coords(seg2[1])

        # vector of the center of seg2 (from the origin)
        C = 0.5 * (s1 + s2)

        # endpoint of seg1 (from the origin)
        P = coords(seg1[1])

        # vector from the center C of seg2 to the endpoint P of seg1
        CP = P - C

        # vector of seg2
        S1S2 = s2 - s1

        # projection of CP upon seg2
        prj = mm.vector_projection(CP, S1S2)

        # check if the distance of the orthogonal complement of CP projection on S1S2
        # (vertical distance from P to seg2) is smaller than the sum of the radii. (overlap)
        # If not exit early, because there is no way that backtracking can feasible
        if not is_seg2_within_seg1_radius(np.linalg.norm(CP - prj), seg1, seg2):
            return False

        # projection lies within the length of the cylinder. Check if the distance between
        # the center C of seg2 and the projection of the end point of seg1, P is smaller than
        # half of the others length plus a 5% tolerance
        return np.linalg.norm(prj) < 0.55 * np.linalg.norm(S1S2)

    def is_inside_cylinder(seg1, seg2):
        """Checks if seg2 approximately lies within a cylindrical volume of seg1.

        Two conditions must be satisfied:
            1. The two segments are not facing the same direction  (seg2 comes back to seg1)
            2. seg2 is overlaping with seg1
        """
        return not is_in_the_same_verse(seg1, seg2) and is_seg1_overlapping_with_seg2(seg1, seg2)

    # filter out single segment sections
    section_itr = (snode for snode in neurite.iter_sections() if snode.points.shape[0] > 2)
    for snode in section_itr:
        # group each section's points intro triplets
        segment_pairs = list(filter(is_not_zero_seg, pair(snode.points)))

        # filter out zero length segments
        for i, seg1 in enumerate(segment_pairs[1:]):
            # check if the end point of the segment lies within the previous
            # ones in the current sectionmake
            for seg2 in segment_pairs[0: i + 1]:
                if is_inside_cylinder(seg1, seg2):
                    return True
    return False


def get_flat_neurites(neuron, tol=0.1, method='ratio'):
    """Check if a neuron has neurites that are flat within a tolerance.

    Args:
        neurite(Neurite): neurite to operate on
        tol(float): the tolerance or the ratio
        method(string): 'tolerance' or 'ratio' described in :meth:`is_flat`

    Returns:
        Bool list corresponding to the flatness check for each neurite
        in neuron neurites with respect to the given criteria
    """
    return [n for n in neuron.neurites if is_flat(n, tol, method)]


def get_back_tracking_neurites(neuron):
    """Get neurites that have back-tracks.

    A back-track is the placement of a point near a previous segment during
    the reconstruction, causing a zigzag jump in the morphology which can
    cause issues with meshing algorithms.

    Args:
        neurite(Neurite): neurite to operate on

    Returns:
        List of neurons with backtracks
    """
    return [n for n in neuron.neurites if is_back_tracking(n)]



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
from pathlib import Path
import warnings

import numpy as np
from nose.tools import (assert_almost_equal, assert_equal,
                        assert_greater_equal, assert_raises, ok_)
import pandas as pd
from pandas.testing import assert_frame_equal

import neurom as nm
from neurom.apps import morph_stats as ms
from neurom.exceptions import ConfigError
from neurom.features import NEURITEFEATURES, NEURONFEATURES


DATA_PATH = Path(__file__).parent.parent.parent.parent / 'test_data'
SWC_PATH = DATA_PATH / 'swc'


REF_CONFIG = {
    'neurite': {
        'section_lengths': ['max', 'total'],
        'section_volumes': ['total'],
        'section_branch_orders': ['max'],
        'segment_midpoints': ['max'],
    },
    'neurite_type': ['AXON', 'APICAL_DENDRITE', 'BASAL_DENDRITE', 'ALL'],
    'neuron': {
        'soma_radii': ['mean'],
    }
}

REF_OUT = {
    'mean_soma_radius': 0.13065629648763766,
    'axon': {
        'total_section_length': 207.87975220908129,
        'max_section_length': 11.018460736176685,
        'max_section_branch_order': 10,
        'total_section_volume': 276.73857657289523,
        'max_segment_midpoint_X': 0.0,
        'max_segment_midpoint_Y': 0.0,
        'max_segment_midpoint_Z': 49.520305964149998,
    },
    'all': {
        'total_section_length': 840.68521442251949,
        'max_section_length': 11.758281556059444,
        'max_section_branch_order': 10,
        'total_section_volume': 1104.9077419665782,
        'max_segment_midpoint_X': 64.401674984050004,
        'max_segment_midpoint_Y': 48.48197694465,
        'max_segment_midpoint_Z': 53.750947521650005,
    },
    'apical_dendrite': {
        'total_section_length': 214.37304577550353,
        'max_section_length': 11.758281556059444,
        'max_section_branch_order': 10,
        'total_section_volume': 271.9412385728449,
        'max_segment_midpoint_X': 64.401674984050004,
        'max_segment_midpoint_Y': 0.0,
        'max_segment_midpoint_Z': 53.750947521650005,
    },
    'basal_dendrite': {
        'total_section_length': 418.43241643793476,
        'max_section_length': 11.652508126101711,
        'max_section_branch_order': 10,
        'total_section_volume': 556.22792682083821,
        'max_segment_midpoint_X': 64.007872333250006,
        'max_segment_midpoint_Y': 48.48197694465,
        'max_segment_midpoint_Z': 51.575580778049996,
    },
}


def test_name_correction():
    assert_equal(ms._stat_name('foo', 'raw'), 'foo')
    assert_equal(ms._stat_name('foos', 'raw'), 'foo')
    assert_equal(ms._stat_name('foos', 'bar'), 'bar_foo')
    assert_equal(ms._stat_name('foos', 'total'), 'total_foo')
    assert_equal(ms._stat_name('soma_radii', 'total'), 'total_soma_radius')
    assert_equal(ms._stat_name('soma_radii', 'raw'), 'soma_radius')


def test_eval_stats_raw_returns_list():
    assert_equal(ms.eval_stats(np.array([1, 2, 3, 4]), 'raw'), [1, 2, 3, 4])


def test_eval_stats_empty_input_returns_none():
    ok_(ms.eval_stats([], 'min') is None)


def test_eval_stats_total_returns_sum():
    assert_equal(ms.eval_stats(np.array([1, 2, 3, 4]), 'total'), 10)


def test_eval_stats_applies_numpy_function():
    modes = ('min', 'max', 'mean', 'median', 'std')

    ref_array = np.arange(1, 10)

    for m in modes:
        assert_equal(ms.eval_stats(ref_array, m),
            getattr(np, m)(ref_array))


def test_extract_stats_single_neuron():
    nrn = nm.load_neuron(Path(SWC_PATH, 'Neuron.swc'))
    res = ms.extract_stats(nrn, REF_CONFIG)
    assert_equal(set(res.keys()), set(REF_OUT.keys()))
    # Note: soma radius is calculated from the sphere that gives the area
    # of the cylinders described in Neuron.swc
    assert_almost_equal(res['mean_soma_radius'], REF_OUT['mean_soma_radius'])

    for k in ('all', 'axon', 'basal_dendrite', 'apical_dendrite'):
        assert_equal(set(res[k].keys()), set(REF_OUT[k].keys()))
        for kk in res[k].keys():
            assert_almost_equal(res[k][kk], REF_OUT[k][kk], places=3)


def test_extract_dataframe():
    # Vanilla test
    nrns = nm.load_neurons([Path(SWC_PATH, name)
                            for name in ['Neuron.swc', 'simple.swc']])
    actual = ms.extract_dataframe(nrns, REF_CONFIG)
    expected = pd.read_csv(Path(DATA_PATH, 'extracted-stats.csv'), index_col=0)
    assert_frame_equal(actual, expected)

    # Test with a single neuron in the population
    nrns = nm.load_neurons(Path(SWC_PATH, 'Neuron.swc'))
    actual = ms.extract_dataframe(nrns, REF_CONFIG)
    assert_frame_equal(actual, expected[expected.name == 'Neuron'], check_dtype=False)

    # Test with a config without the 'neuron' key
    nrns = nm.load_neurons([Path(SWC_PATH, name)
                            for name in ['Neuron.swc', 'simple.swc']])
    config = {'neurite': {'section_lengths': ['total']},
              'neurite_type': ['AXON', 'APICAL_DENDRITE', 'BASAL_DENDRITE', 'ALL']}
    actual = ms.extract_dataframe(nrns, config)
    expected = expected[['name', 'neurite_type', 'total_section_length']]
    assert_frame_equal(actual, expected)

    # Test with a FstNeuron argument
    nrn = nm.load_neuron(Path(SWC_PATH, 'Neuron.swc'))
    actual = ms.extract_dataframe(nrn, config)
    assert_frame_equal(actual, expected[expected.name == 'Neuron'], check_dtype=False)

    # Test with a List[FstNeuron] argument
    nrns = [nm.load_neuron(Path(SWC_PATH, name))
            for name in ['Neuron.swc', 'simple.swc']]
    actual = ms.extract_dataframe(nrns, config)
    assert_frame_equal(actual, expected)

    # Test with a List[Path] argument
    nrns = [Path(SWC_PATH, name) for name in ['Neuron.swc', 'simple.swc']]
    actual = ms.extract_dataframe(nrns, config)
    assert_frame_equal(actual, expected)

    # Test without any neurite_type keys, it should pick the defaults
    config = {'neurite': {'total_length_per_neurite': ['total']}}
    actual = ms.extract_dataframe(nrns, config)
    expected = pd.DataFrame(
        columns=['name', 'neurite_type', 'total_total_length_per_neurite'],
        data=[['Neuron', 'axon', 207.879752],
              ['Neuron', 'basal_dendrite', 418.432416],
              ['Neuron', 'apical_dendrite', 214.373046],
              ['Neuron', 'all', 840.685214],
              ['simple', 'axon', 15.000000],
              ['simple', 'basal_dendrite', 16.000000],
              ['simple', 'apical_dendrite', 0.000000],
              ['simple', 'all', 31.000000]])
    assert_frame_equal(actual, expected)


# def test_extract_dataframe_multiproc():
#     nrns = nm.load_neurons([Path(SWC_PATH, name)
#                             for name in ['Neuron.swc', 'simple.swc']])
#     with warnings.catch_warnings(record=True) as w:
#         actual = ms.extract_dataframe(nrns, REF_CONFIG, n_workers=2)
#     expected = pd.read_csv(Path(DATA_PATH, 'extracted-stats.csv'), index_col=0)

#     # Compare sorted DataFrame since Pool.imap_unordered disrupted the order
#     assert_frame_equal(actual.sort_values(by=['name']).reset_index(drop=True),
#                        expected.sort_values(by=['name']).reset_index(drop=True))

#     with warnings.catch_warnings(record=True) as w:
#         actual = ms.extract_dataframe(nrns, REF_CONFIG, n_workers=os.cpu_count() + 1)
#         assert_equal(len(w), 1, "Warning not emitted")
#     assert_frame_equal(actual.sort_values(by=['name']).reset_index(drop=True),
#                        expected.sort_values(by=['name']).reset_index(drop=True))



def test_get_header():
    fake_results = {'fake_name0': REF_OUT,
                    'fake_name1': REF_OUT,
                    'fake_name2': REF_OUT,
                    }
    header = ms.get_header(fake_results)
    assert_equal(1 + 1 + 4 * (4 + 3), len(header))  # name + everything in REF_OUT
    ok_('name' in header)
    ok_('mean_soma_radius' in header)


def test_generate_flattened_dict():
    fake_results = {'fake_name0': REF_OUT,
                    'fake_name1': REF_OUT,
                    'fake_name2': REF_OUT,
                    }
    header = ms.get_header(fake_results)
    rows = list(ms.generate_flattened_dict(header, fake_results))
    assert_equal(3, len(rows))  # one for fake_name[0-2]
    assert_equal(1 + 1 + 4 * (4 + 3), len(rows[0]))  # name + everything in REF_OUT


def test_full_config():
    config = ms.full_config()
    assert_equal(set(config.keys()), {'neurite', 'neuron', 'neurite_type'})

    assert_equal(set(config['neurite'].keys()), set(NEURITEFEATURES.keys()))
    assert_equal(set(config['neuron'].keys()), set(NEURONFEATURES.keys()))


def test_sanitize_config():
    assert_raises(ConfigError, ms.sanitize_config, {'neurite': []})

    new_config = ms.sanitize_config({})  # empty
    assert_equal(2, len(new_config))  # neurite & neuron created

    full_config = {
        'neurite': {
            'section_lengths': ['max', 'total'],
            'section_volumes': ['total'],
            'section_branch_orders': ['max']
        },
        'neurite_type': ['AXON', 'APICAL_DENDRITE', 'BASAL_DENDRITE', 'ALL'],
        'neuron': {
            'soma_radii': ['mean']
        }
    }
    new_config = ms.sanitize_config(full_config)
    assert_equal(3, len(new_config))  # neurite, neurite_type & neuron



# # Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# # All rights reserved.
# #
# # This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
# #
# # Redistribution and use in source and binary forms, with or without
# # modification, are permitted provided that the following conditions are met:
# #
# #     1. Redistributions of source code must retain the above copyright
# #        notice, this list of conditions and the following disclaimer.
# #     2. Redistributions in binary form must reproduce the above copyright
# #        notice, this list of conditions and the following disclaimer in the
# #        documentation and/or other materials provided with the distribution.
# #     3. Neither the name of the copyright holder nor the names of
# #        its contributors may be used to endorse or promote products
# #        derived from this software without specific prior written permission.
# #
# # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# # ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# # WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# # DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# # DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# # (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# # LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# # ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# # (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# # SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# import math

# import numpy as np
# from nose import tools as nt

# import neurom as nm
# from neurom.core import Neurite, Section

# RADIUS = 4.
# POINTS0 = np.array([[0., 0., 0., RADIUS],
#                     [0., 0., 1., RADIUS],
#                     [0., 0., 2., RADIUS],
#                     [0., 0., 3., RADIUS],
#                     [1., 0., 3., RADIUS],
#                     [2., 0., 3., RADIUS],
#                     [3., 0., 3., RADIUS]])

# POINTS1 = np.array([[3., 0., 3., RADIUS],
#                     [3., 0., 4., RADIUS],
#                     [3., 0., 5., RADIUS],
#                     [3., 0., 6., RADIUS],
#                     [4., 0., 6., RADIUS],
#                     [5., 0., 6., RADIUS],
#                     [6., 0., 6., RADIUS]])

# REF_LEN = 12


# ROOT_NODE = Section(POINTS0)
# ROOT_NODE.add_child(Section(POINTS1))


# def test_init():
#     nrt = Neurite(ROOT_NODE)
#     nt.eq_(nrt.type, nm.NeuriteType.undefined)
#     nt.eq_(len(nrt.points), 13)


# def test_neurite_type():
#     root_node = Section(POINTS0, section_type=nm.AXON)
#     nrt = Neurite(root_node)
#     nt.eq_(nrt.type, nm.AXON)

#     root_node = Section(POINTS0, section_type=nm.BASAL_DENDRITE)
#     nrt = Neurite(root_node)
#     nt.eq_(nrt.type, nm.BASAL_DENDRITE)


# def test_neurite_length():
#     nrt = Neurite(ROOT_NODE)
#     nt.assert_almost_equal(nrt.length, REF_LEN)


# def test_neurite_area():
#     nrt = Neurite(ROOT_NODE)
#     area = 2 * math.pi * RADIUS * REF_LEN
#     nt.assert_almost_equal(nrt.area, area)


# def test_neurite_volume():
#     nrt = Neurite(ROOT_NODE)
#     volume = math.pi * RADIUS * RADIUS * REF_LEN
#     nt.assert_almost_equal(nrt.volume, volume)


# def test_str():
#     nrt = Neurite(ROOT_NODE)
#     nt.ok_('Neurite' in str(nrt))


# def test_neurite_hash():
#     nrt = Neurite(ROOT_NODE)
#     nt.eq_(hash(nrt), hash((nrt.type, nrt.root_node)))




# Copyright (c) 2016, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from mock import Mock
from nose import tools as nt

from neurom.core.types import (NEURITES, NeuriteType, axon_filter,
                               dendrite_filter, tree_type_checker)


def test_tree_type_checker():
    # check that when NeuriteType.all, we accept all trees, w/o checking type
    tree_filter = tree_type_checker(NeuriteType.all)
    nt.ok_(tree_filter('fake_tree'))

    mock_tree = Mock()
    mock_tree.type = NeuriteType.axon

    # single arg
    tree_filter = tree_type_checker(NeuriteType.axon)
    nt.ok_(tree_filter(mock_tree))

    mock_tree.type = NeuriteType.basal_dendrite
    nt.ok_(not tree_filter(mock_tree))

    # multiple args
    tree_filter = tree_type_checker(NeuriteType.axon, NeuriteType.basal_dendrite)
    nt.ok_(tree_filter(mock_tree))

    tree_filter = tree_type_checker(*NEURITES)
    nt.ok_(tree_filter(mock_tree))


# FIXME: https://github.com/BlueBrain/NeuroM/issues/804
@nt.nottest
def test_tree_type_checker_broken():
    tree_filter = tree_type_checker(NeuriteType.all)
    mock_tree = Mock()
    mock_tree.type = NeuriteType.soma
    nt.ok_(not tree_filter(mock_tree))

    mock_tree.type = NeuriteType.undefined
    nt.ok_(not tree_filter(mock_tree))


def test_type_filters():
    axon = Mock()
    axon.type = NeuriteType.axon
    nt.ok_(axon_filter(axon))
    nt.ok_(not dendrite_filter(axon))

    basal_dendrite = Mock()
    basal_dendrite.type = NeuriteType.basal_dendrite
    nt.ok_(not axon_filter(basal_dendrite))
    nt.ok_(dendrite_filter(basal_dendrite))

    apical_dendrite = Mock()
    apical_dendrite.type = NeuriteType.apical_dendrite
    nt.ok_(not axon_filter(apical_dendrite))
    nt.ok_(dendrite_filter(apical_dendrite))



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from .utils import get_fig_2d, get_fig_3d # needs to be at top to trigger matplotlib Agg backend

from pathlib import Path
from nose import tools as nt
import shutil
import tempfile

import numpy as np


from neurom.view.common import (plt, figure_naming, get_figure, save_plot, plot_style,
                                plot_title, plot_labels, plot_legend, update_plot_limits, plot_ticks,
                                plot_sphere, plot_cylinder)


def test_figure_naming():
    pretitle, posttitle, prefile, postfile = figure_naming(pretitle='Test', prefile="", postfile=3)
    nt.eq_(pretitle, 'Test -- ')
    nt.eq_(posttitle, "")
    nt.eq_(prefile, "")
    nt.eq_(postfile, "_3")

    pretitle, posttitle, prefile, postfile = figure_naming(pretitle='', posttitle="Test", prefile="test", postfile="")
    nt.eq_(pretitle, "")
    nt.eq_(posttitle, " -- Test")
    nt.eq_(prefile, "test_")
    nt.eq_(postfile, "")


def test_get_figure():
    fig_old = plt.figure()
    fig, ax = get_figure(new_fig=False)
    nt.eq_(fig, fig_old)
    nt.eq_(ax.get_subplotspec().colspan.start, 0)
    nt.eq_(ax.get_subplotspec().rowspan.start, 0)

    fig1, ax1 = get_figure(new_fig=True, subplot=224)
    nt.ok_(fig1 != fig_old)
    nt.eq_(ax1.get_subplotspec().colspan.start, 1)
    nt.eq_(ax1.get_subplotspec().rowspan.start, 1)

    fig2, ax2 = get_figure(new_fig=True, subplot=[1, 1, 1])
    nt.eq_(ax2.get_subplotspec().colspan.start, 0)
    nt.eq_(ax2.get_subplotspec().rowspan.start, 0)
    plt.close('all')

    fig = plt.figure()
    ax = fig.add_subplot(111)
    fig2, ax2 = get_figure(new_fig=False)
    nt.eq_(fig2, plt.gcf())
    nt.eq_(ax2, plt.gca())
    plt.close('all')


def test_save_plot():
    with tempfile.TemporaryDirectory() as folder:
        fig_old = plt.figure()
        output_path = Path(folder, 'subdir')
        fig = save_plot(fig_old, output_path=output_path)
        nt.ok_(Path(output_path, 'Figure.png').is_file())
        plt.close('all')



def test_plot_title():
    with get_fig_2d() as (fig, ax):
        plot_title(ax)
        nt.eq_(ax.get_title(), 'Figure')

    with get_fig_2d() as (fig, ax):
        plot_title(ax, title='Test')
        nt.eq_(ax.get_title(), 'Test')


def test_plot_labels():
    with get_fig_2d() as (fig, ax):
        plot_labels(ax)
        nt.eq_(ax.get_xlabel(), 'X')
        nt.eq_(ax.get_ylabel(), 'Y')

    with get_fig_2d() as (fig, ax):
        plot_labels(ax, xlabel='T', ylabel='R')
        nt.eq_(ax.get_xlabel(), 'T')
        nt.eq_(ax.get_ylabel(), 'R')

    with get_fig_3d() as (fig0, ax0):
        plot_labels(ax0)
        nt.eq_(ax0.get_zlabel(), 'Z')

    with get_fig_3d() as (fig0, ax0):
        plot_labels(ax0, zlabel='T')
        nt.eq_(ax0.get_zlabel(), 'T')


def test_plot_legend():
    with get_fig_2d() as (fig, ax):
        plot_legend(ax)
        legend = ax.get_legend()
        nt.ok_(legend is None)

    with get_fig_2d() as (fig, ax):
        ax.plot([1, 2, 3], [1, 2, 3], label='line 1')
        plot_legend(ax, no_legend=False)
        legend = ax.get_legend()
        nt.eq_(legend.get_texts()[0].get_text(), 'line 1')


def test_plot_limits():
    with get_fig_2d() as (fig, ax):
        nt.assert_raises(AssertionError, update_plot_limits, ax, white_space=0)

    with get_fig_2d() as (fig, ax):
        ax.dataLim.update_from_data_xy(((0, -100), (100, 0)))

        update_plot_limits(ax, white_space=0)
        nt.eq_(ax.get_xlim(), (0, 100))
        nt.eq_(ax.get_ylim(), (-100, 0))

    with get_fig_3d() as (fig0, ax0):
        update_plot_limits(ax0, white_space=0)
        zlim0 = ax0.get_zlim()
        nt.ok_(np.allclose(ax0.get_zlim(), zlim0))


def test_plot_ticks():
    with get_fig_2d() as (fig, ax):
        plot_ticks(ax)
        nt.ok_(len(ax.get_xticks()))
        nt.ok_(len(ax.get_yticks()))

    with get_fig_2d() as (fig, ax):
        plot_ticks(ax, xticks=[], yticks=[])
        nt.eq_(len(ax.get_xticks()), 0)
        nt.eq_(len(ax.get_yticks()), 0)

    with get_fig_2d() as (fig, ax):
        plot_ticks(ax, xticks=np.arange(3), yticks=np.arange(4))
        nt.eq_(len(ax.get_xticks()), 3)
        nt.eq_(len(ax.get_yticks()), 4)

    with get_fig_3d() as (fig0, ax0):
        plot_ticks(ax0)
        nt.ok_(len(ax0.get_zticks()))

    with get_fig_3d() as (fig0, ax0):
        plot_ticks(ax0, zticks=[])
        nt.eq_(len(ax0.get_zticks()), 0)

    with get_fig_3d() as (fig0, ax0):
        plot_ticks(ax0, zticks=np.arange(3))
        nt.eq_(len(ax0.get_zticks()), 3)


def test_plot_style():
    with get_fig_2d() as (fig, ax):
        ax.dataLim.update_from_data_xy(((0, -100), (100, 0)))

        plot_style(fig, ax)

        nt.eq_(ax.get_title(), 'Figure')
        nt.eq_(ax.get_xlabel(), 'X')
        nt.eq_(ax.get_ylabel(), 'Y')

    with get_fig_2d() as (fig, ax):
        ax.dataLim.update_from_data_xy(((0, -100), (100, 0)))

        plot_style(fig, ax, no_axes=True)

        nt.ok_(not ax.get_frame_on())
        nt.ok_(not ax.xaxis.get_visible())
        nt.ok_(not ax.yaxis.get_visible())

    with get_fig_2d() as (fig, ax):
        ax.dataLim.update_from_data_xy(((0, -100), (100, 0)))
        plot_style(fig, ax, tight=True)
        nt.ok_(fig.get_tight_layout())


def test_plot_cylinder():
    fig0, ax0 = get_figure(params={'projection': '3d'})
    start, end = np.array([0, 0, 0]), np.array([1, 0, 0])
    plot_cylinder(ax0, start=start, end=end,
                  start_radius=0, end_radius=10.,
                  color='black', alpha=1.)
    nt.ok_(ax0.has_data())


def test_plot_sphere():
    fig0, ax0 = get_figure(params={'projection': '3d'})
    plot_sphere(ax0, [0, 0, 0], 10., color='black', alpha=1.)
    nt.ok_(ax0.has_data())



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pathlib import Path
from io import StringIO
from itertools import chain

from nose import tools as nt
from numpy.testing import assert_array_equal

import neurom as nm
from neurom import COLS, core, load_neuron
from neurom.core import NeuriteIter, Tree

DATA_PATH = Path(__file__).parent.parent.parent.parent / 'test_data'

NRN1 = load_neuron(Path(DATA_PATH, 'swc/Neuron.swc'))

NEURONS = [NRN1,
           load_neuron(Path(DATA_PATH, 'swc/Single_basal.swc')),
           load_neuron(Path(DATA_PATH, 'swc/Neuron_small_radius.swc')),
           load_neuron(Path(DATA_PATH, 'swc/Neuron_3_random_walker_branches.swc')),
           ]
TOT_NEURITES = sum(len(N.neurites) for N in NEURONS)

SIMPLE = load_neuron(Path(DATA_PATH, 'swc/simple.swc'))
REVERSED_NEURITES = load_neuron(Path(DATA_PATH, 'swc/ordering/reversed_NRN_neurite_order.swc'))

POP = core.Population(NEURONS, name='foo')


def assert_sequence_equal(a, b):
    nt.eq_(tuple(a), tuple(b))


def test_iter_neurites_default():
    assert_sequence_equal(POP.neurites,
                          [n for n in core.iter_neurites(POP)])


def test_iter_neurites_nrn_order():
    assert_sequence_equal(list(core.iter_neurites(REVERSED_NEURITES,
                                                  neurite_order=NeuriteIter.NRN)),
                          reversed(list(core.iter_neurites(REVERSED_NEURITES))))


def test_iter_neurites_filter():

    for ntyp in nm.NEURITE_TYPES:
        a = [n for n in POP.neurites if n.type == ntyp]
        b = [n for n in core.iter_neurites(POP, filt=lambda n: n.type == ntyp)]
        assert_sequence_equal(a, b)


def test_iter_neurites_mapping():

    n = [n for n in core.iter_neurites(POP, mapfun=lambda n: len(n.points))]
    ref = [211, 211, 211, 211, 211, 211, 211, 211, 211, 500, 500, 500]
    assert_sequence_equal(n, ref)


def test_iter_neurites_filter_mapping():
    n = [n for n in core.iter_neurites(POP,
                                       mapfun=lambda n: len(n.points),
                                       filt=lambda n: len(n.points) > 250)]

    ref = [500, 500, 500]
    assert_sequence_equal(n, ref)


def test_iter_sections_default():

    ref = [s for n in POP.neurites for s in n.iter_sections()]
    assert_sequence_equal(ref,
                          [n for n in core.iter_sections(POP)])

def test_iter_sections_default_pop():
    ref = [s.id for n in POP.neurites for s in n.iter_sections()]
    assert_sequence_equal(ref, [n.id for n in core.iter_sections(POP)])


def test_iter_sections_filter():
    for ntyp in nm.NEURITE_TYPES:
        a = [s.id for n in filter(lambda nn: nn.type == ntyp, POP.neurites)
             for s in n.iter_sections()]
        b = [n.id for n in core.iter_sections(POP, neurite_filter=lambda n: n.type == ntyp)]
        assert_sequence_equal(a, b)

def test_iter_sections_inrnorder():
    # The + 2 comes from the fact the sequence used to be offset by 2 (only in the case of 3 point soma) in NeuroM v1
    # TODO: Cleans this up in a second step once Neurom v2 has been merged to master
    assert_sequence_equal([s.id + 2 for n in POP.neurites for s in n.iter_sections(neurite_order=NeuriteIter.NRN)],
                          [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 2, 3, 4])

def test_iter_sections_ipreorder():
    assert_sequence_equal([s.id + 2 for n in POP.neurites for s in n.iter_sections(Tree.ipreorder)],
                          [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 2, 3, 4])

def test_iter_sections_ipostorder():
    assert_sequence_equal([s.id + 2 for n in POP.neurites for s in n.iter_sections(Tree.ipostorder)],
                          [3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 22, 20, 18, 16, 14, 12, 10, 8, 6, 4, 2, 24, 26, 28, 30, 32, 34, 36, 38, 40, 42, 43, 41, 39, 37, 35, 33, 31, 29, 27, 25, 23, 45, 47, 49, 51, 53, 55, 57, 59, 61, 63, 64, 62, 60, 58, 56, 54, 52, 50, 48, 46, 44, 66, 68, 70, 72, 74, 76, 78, 80, 82, 84, 85, 83, 81, 79, 77, 75, 73, 71, 69, 67, 65, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 22, 20, 18, 16, 14, 12, 10, 8, 6, 4, 2, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 22, 20, 18, 16, 14, 12, 10, 8, 6, 4, 2, 24, 26, 28, 30, 32, 34, 36, 38, 40, 42, 43, 41, 39, 37, 35, 33, 31, 29, 27, 25, 23, 45, 47, 49, 51, 53, 55, 57, 59, 61, 63, 64, 62, 60, 58, 56, 54, 52, 50, 48, 46, 44, 66, 68, 70, 72, 74, 76, 78, 80, 82, 84, 85, 83, 81, 79, 77, 75, 73, 71, 69, 67, 65, 2, 3, 4])


def test_iter_sections_ibifurcation():
    assert_sequence_equal([s.id + 2 for n in POP.neurites for s in n.iter_sections(Tree.ibifurcation_point)],
                          [2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 23, 25, 27, 29, 31, 33, 35, 37, 39, 41, 44, 46, 48, 50, 52, 54, 56, 58, 60, 62, 65, 67, 69, 71, 73, 75, 77, 79, 81, 83, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 23, 25, 27, 29, 31, 33, 35, 37, 39, 41, 44, 46, 48, 50, 52, 54, 56, 58, 60, 62, 65, 67, 69, 71, 73, 75, 77, 79, 81, 83],)


def test_iter_sections_iforking():
    assert_sequence_equal([s.id + 2 for n in POP.neurites for s in n.iter_sections(Tree.iforking_point)],
                          [2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 23, 25, 27, 29, 31, 33, 35, 37, 39, 41, 44, 46, 48, 50, 52, 54, 56, 58, 60, 62, 65, 67, 69, 71, 73, 75, 77, 79, 81, 83, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 23, 25, 27, 29, 31, 33, 35, 37, 39, 41, 44, 46, 48, 50, 52, 54, 56, 58, 60, 62, 65, 67, 69, 71, 73, 75, 77, 79, 81, 83])


def test_iter_sections_ileaf():
    assert_sequence_equal([s.id + 2 for n in POP.neurites for s in n.iter_sections(Tree.ileaf)],
                          [3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 22, 24, 26, 28, 30, 32, 34, 36, 38, 40, 42, 43, 45, 47, 49, 51, 53, 55, 57, 59, 61, 63, 64, 66, 68, 70, 72, 74, 76, 78, 80, 82, 84, 85, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 22, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 22, 24, 26, 28, 30, 32, 34, 36, 38, 40, 42, 43, 45, 47, 49, 51, 53, 55, 57, 59, 61, 63, 64, 66, 68, 70, 72, 74, 76, 78, 80, 82, 84, 85, 2, 3, 4])


def test_iter_section_nrn():
    ref = list(core.iter_sections(SIMPLE))
    nt.eq_(len(ref), 6)

    ref = list(core.iter_sections(SIMPLE, neurite_filter=lambda n: n.type == nm.AXON))
    nt.eq_(len(ref), 3)

    ref = list(core.iter_sections(SIMPLE, neurite_filter=lambda n: n.type == nm.BASAL_DENDRITE))
    nt.eq_(len(ref), 3)

    ref = list(core.iter_sections(SIMPLE, neurite_filter=lambda n: n.type == nm.APICAL_DENDRITE))
    nt.eq_(len(ref), 0)


def test_iter_segments_nrn():
    ref = list(core.iter_segments(SIMPLE))
    nt.eq_(len(ref), 6)

    ref = list(core.iter_segments(SIMPLE, neurite_filter=lambda n: n.type == nm.AXON))
    nt.eq_(len(ref), 3)

    ref = list(core.iter_segments(SIMPLE, neurite_filter=lambda n: n.type == nm.BASAL_DENDRITE))
    nt.eq_(len(ref), 3)

    ref = list(core.iter_segments(SIMPLE, neurite_filter=lambda n: n.type == nm.APICAL_DENDRITE))
    nt.eq_(len(ref), 0)

    ref = list(core.iter_segments(NRN1))
    nt.eq_(len(ref), 840)

    ref = list(core.iter_segments(NRN1, neurite_filter=lambda n: n.type == nm.AXON))
    nt.eq_(len(ref), 210)

    ref = list(core.iter_segments(NRN1, neurite_filter=lambda n: n.type == nm.BASAL_DENDRITE))
    nt.eq_(len(ref), 420)

    ref = list(core.iter_segments(NRN1, neurite_filter=lambda n: n.type == nm.APICAL_DENDRITE))
    nt.eq_(len(ref), 210)


def test_iter_segments_pop():

    ref = list(core.iter_segments(POP))
    nt.eq_(len(ref), 3387)

    ref = list(core.iter_segments(POP, neurite_filter=lambda n: n.type == nm.AXON))
    nt.eq_(len(ref), 919)

    ref = list(core.iter_segments(POP, neurite_filter=lambda n: n.type == nm.BASAL_DENDRITE))
    nt.eq_(len(ref), 1549)

    ref = list(core.iter_segments(POP, neurite_filter=lambda n: n.type == nm.APICAL_DENDRITE))
    nt.eq_(len(ref), 919)


def test_iter_segments_section():
    sec = load_neuron(StringIO(u"""
	                      ((CellBody)
	                       (0 0 0 2))

	                      ((Dendrite)
                          (1 2 3 8)
                          (5 6 7 16)
                          (8 7 6 10)
                          (4 3 2 2))
                       """), reader='asc').sections[0]
    ref = [[p1[COLS.XYZR].tolist(), p2[COLS.XYZR].tolist()]
           for p1, p2 in core.iter_segments(sec)]

    assert_array_equal(ref, [[[1, 2, 3, 4], [5, 6, 7, 8]],
                             [[5, 6, 7, 8], [8, 7, 6, 5]],
                             [[8, 7, 6, 5], [4, 3, 2, 1]]])



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
from pathlib import Path
import shutil
import tempfile
import mock
import sys

import matplotlib
if 'DISPLAY' not in os.environ:  # noqa
    matplotlib.use('Agg')  # noqa

import neurom
from neurom.view import common, plotly
from neurom import load_neuron, viewer, NeuriteType

from nose import tools as nt
from numpy.testing import assert_allclose, assert_array_almost_equal

DATA_PATH = Path(__file__).parent.parent.parent / 'test_data/swc'
MORPH_FILENAME = Path(DATA_PATH, 'Neuron.swc')

nrn = load_neuron(MORPH_FILENAME)


def _reload_module(module):
    """Force module reload."""
    import importlib
    importlib.reload(module)


def test_plotly_extra_not_installed():
    with mock.patch.dict(sys.modules, {'plotly': None}):
        try:
            _reload_module(neurom.view.plotly)
            nt.ok_(False, "ImportError not triggered")
        except ImportError as e:
            nt.assert_equal(str(e),
                            'neurom[plotly] is not installed. '
                            'Please install it by doing: pip install neurom[plotly]')


def test_plotly_draw_neuron3d():
    plotly.draw(nrn, plane='3d', auto_open=False)
    plotly.draw(nrn.neurites[0], plane='3d', auto_open=False)

    fig = plotly.draw(load_neuron(Path(DATA_PATH, 'simple-different-soma.swc')),
                      auto_open=False)
    x, y, z = [fig['data'][2][key] for key in str('xyz')]
    assert_allclose(x[0, 0], 2)
    assert_allclose(x[33, 33], -1.8971143170299758)
    assert_allclose(y[0, 0], 3)
    assert_allclose(y[33, 33], 9.75)
    assert_allclose(z[0, 0], 13)
    assert_allclose(z[33, 33], 8.5)

def test_plotly_draw_neuron2d():
    plotly.draw(nrn, plane='xy', auto_open=False)
    plotly.draw(nrn.neurites[0], plane='xy', auto_open=False)


def test_draw_neuron():
    viewer.draw(nrn)
    common.plt.close('all')


def test_draw_filter_neurite():
    for mode in ['2d', '3d']:
        viewer.draw(nrn, mode=mode, neurite_type=NeuriteType.basal_dendrite)
        assert_allclose(common.plt.gca().get_ylim(),
                        [-30., 78], atol=5)

    common.plt.close('all')


def test_draw_neuron3d():
    viewer.draw(nrn, mode='3d')
    common.plt.close('all')

    nt.assert_raises(NotImplementedError, viewer.draw, nrn, mode='3d', realistic_diameters=True)


def test_draw_tree():
    viewer.draw(nrn.neurites[0])
    common.plt.close('all')


def test_draw_tree3d():
    viewer.draw(nrn.neurites[0], mode='3d')
    common.plt.close('all')


def test_draw_soma():
    viewer.draw(nrn.soma)
    common.plt.close('all')


def test_draw_soma3d():
    viewer.draw(nrn.soma, mode='3d')
    common.plt.close('all')


def test_draw_dendrogram():
    viewer.draw(nrn, mode='dendrogram')
    common.plt.close('all')

    viewer.draw(nrn.neurites[0], mode='dendrogram')
    common.plt.close('all')

def test_draw_dendrogram_empty_segment():
    neuron = load_neuron(Path(DATA_PATH, 'empty_segments.swc'))
    viewer.draw(neuron, mode='dendrogram')
    common.plt.close('all')



@nt.raises(viewer.InvalidDrawModeError)
def test_invalid_draw_mode_raises():
    viewer.draw(nrn, mode='4d')


@nt.raises(viewer.NotDrawableError)
def test_invalid_object_raises():
    class Dummy(object):
        pass
    viewer.draw(Dummy())


@nt.raises(viewer.NotDrawableError)
def test_invalid_combo_raises():
    viewer.draw(nrn.soma, mode='dendrogram')


def test_writing_output():
    with tempfile.TemporaryDirectory() as folder:
        output_dir = Path(folder, 'subdir')
        viewer.draw(nrn, mode='2d', output_path=output_dir)
        nt.ok_((output_dir / 'Figure.png').is_file())
        common.plt.close('all')



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test neurom.features_get features."""

import math
from io import StringIO
from pathlib import Path

import neurom as nm
import numpy as np
from neurom import (core, features, iter_neurites, iter_sections, load_neuron,
                    load_neurons)
from neurom.core.population import Population
from neurom.core.types import NeuriteType
from neurom.core.types import tree_type_checker as _is_type
from neurom.exceptions import NeuroMError
from neurom.features import NEURITEFEATURES
from neurom.features import get as get_feature
from neurom.features import neuritefunc as nf
from nose import tools as nt
from numpy.testing import assert_allclose

DATA_PATH = Path(__file__).parent / '../../../test_data'
NRN_FILES = [Path(DATA_PATH, 'h5/v1', f)
             for f in ('Neuron.h5', 'Neuron_2_branch.h5', 'bio_neuron-001.h5')]

NRNS = load_neurons(NRN_FILES)
NRN = NRNS[0]
POP = Population(NRNS)

SWC_PATH = Path(DATA_PATH, 'swc')
NEURON_PATH = Path(SWC_PATH, 'Neuron.swc')
NEURON = load_neuron(NEURON_PATH)

NEURITES = (NeuriteType.axon,
            NeuriteType.apical_dendrite,
            NeuriteType.basal_dendrite,
            NeuriteType.all)


def assert_items_equal(a, b):
    nt.eq_(sorted(a), sorted(b))


def assert_features_for_neurite(feat, neurons, expected, exact=True):
    for neurite_type, expected_values in expected.items():
        if neurite_type is None:
            res_pop = get_feature(feat, neurons)
            res = get_feature(feat, neurons[0])
        else:
            res_pop = get_feature(feat, neurons, neurite_type=neurite_type)
            res = get_feature(feat, neurons[0], neurite_type=neurite_type)

        if exact:
            assert_items_equal(res_pop, expected_values)
        else:
            assert_allclose(res_pop, expected_values)

        # test for single neuron
        if isinstance(res, np.ndarray):
            # some features, ex: total_length return arrays w/ one element when
            # called on a single neuron
            nt.eq_(len(res), 1)
            res = res[0]
        if exact:
            nt.eq_(res, expected_values[0])
        else:
            assert_allclose(res, expected_values[0])


def _stats(seq):
    return np.min(seq), np.max(seq), np.sum(seq), np.mean(seq)


def test_number_of_sections():
    feat = 'number_of_sections'
    expected = {None: [84, 42, 201],
                NeuriteType.all: [84, 42, 201],
                NeuriteType.axon: [21, 21, 178],
                NeuriteType.apical_dendrite: [21, 0, 0],
                NeuriteType.basal_dendrite: [42, 21, 23],
                }
    assert_features_for_neurite(feat, POP, expected)


def test_section_tortuosity_pop():

    feat = 'section_tortuosity'

    assert_allclose(_stats(get_feature(feat, POP)),
                    [1.0,
                     4.6571118550276704,
                     439.33173,
                     1.343522])

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.all)),
                    (1.0,
                     4.6571118550276704,
                     439.33173,
                     1.343522))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.apical_dendrite)),
                    (1.0702760052031615,
                     1.5732825321954913,
                     26.919574286670883,
                     1.2818844898414707))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.basal_dendrite)),
                    (1.042614577410971,
                     1.6742599832295344,
                     106.5960839640893,
                     1.239489348419643), rtol=1e-5)


def test_section_tortuosity_nrn():
    feat = 'section_tortuosity'
    assert_allclose(_stats(get_feature(feat, NRN)),
                    (1.0702760052031612,
                     1.5732825321954911,
                     106.42449427885093,
                     1.2669582652244158))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.all)),
                    (1.0702760052031612,
                     1.5732825321954911,
                     106.42449427885093,
                     1.2669582652244158))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.apical_dendrite)),
                    (1.0702760052031612,
                     1.5732825321954911,
                     26.919574286670883,
                     1.2818844898414707))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.basal_dendrite)),
                    (1.0788578286296124,
                     1.5504287518256337,
                     51.540901640170489,
                     1.227164324765964))


def test_number_of_segments():

    feat = 'number_of_segments'

    expected = {None: [840, 419, 5179],
                NeuriteType.all: [840, 419, 5179],
                NeuriteType.axon: [210, 209, 4508],
                NeuriteType.apical_dendrite: [210, 0, 0],
                NeuriteType.basal_dendrite: [420, 210, 671],
                }

    assert_features_for_neurite(feat, POP, expected)


def test_number_of_neurites_pop():
    feat = 'number_of_neurites'
    expected = {None: [4, 2, 4],
                NeuriteType.all: [4, 2, 4],
                NeuriteType.axon: [1, 1, 1],
                NeuriteType.apical_dendrite: [1, 0, 0],
                NeuriteType.basal_dendrite: [2, 1, 3],
                }
    assert_features_for_neurite(feat, POP, expected)


def test_number_of_bifurcations_pop():
    feat = 'number_of_bifurcations'
    expected = {None: [40, 20, 97],
                NeuriteType.all: [40, 20, 97],
                NeuriteType.axon: [10, 10, 87],
                NeuriteType.apical_dendrite: [10, 0, 0],
                NeuriteType.basal_dendrite: [20, 10, 10],
                }
    assert_features_for_neurite(feat, POP, expected)


def test_number_of_forking_points_pop():

    feat = 'number_of_forking_points'

    expected = {None: [40, 20, 98],
                NeuriteType.all: [40, 20, 98],
                NeuriteType.axon: [10, 10, 88],
                NeuriteType.apical_dendrite: [10, 0, 0],
                NeuriteType.basal_dendrite: [20, 10, 10],
                }
    assert_features_for_neurite(feat, POP, expected)


def test_number_of_terminations_pop():
    feat = 'number_of_terminations'
    expected = {None: [44, 22, 103],
                NeuriteType.all: [44, 22, 103],
                NeuriteType.axon: [11, 11, 90],
                NeuriteType.apical_dendrite: [11, 0, 0],
                NeuriteType.basal_dendrite: [22, 11, 13],
                }
    assert_features_for_neurite(feat, POP, expected)


def test_total_length_pop():
    feat = 'total_length'
    expected = {None: [840.68522362011538, 418.83424432013902, 13250.825773939932],
                NeuriteType.all: [840.68522362011538, 418.83424432013902, 13250.825773939932],
                NeuriteType.axon: [207.8797736031714, 207.81088341560977, 11767.156115224638],
                NeuriteType.apical_dendrite: [214.37302709169489, 0, 0],
                NeuriteType.basal_dendrite: [418.43242292524889, 211.02336090452931, 1483.6696587152967],
                }
    assert_features_for_neurite(feat, POP, expected, exact=False)


def test_segment_radii_pop():

    feat = 'segment_radii'

    assert_allclose(_stats(get_feature(feat, POP)),
                    (0.079999998211860657,
                        1.2150000333786011,
                        1301.9191725363567,
                        0.20222416473071708))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.all)),
                    (0.079999998211860657,
                        1.2150000333786011,
                        1301.9191725363567,
                        0.20222416473071708))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.apical_dendrite)),
                    (0.13142434507608414,
                        1.0343990325927734,
                        123.41135908663273,
                        0.58767313850777492))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.basal_dendrite)),
                    (0.079999998211860657,
                        1.2150000333786011,
                        547.43900821779164,
                        0.42078324997524336))


def test_segment_radii_nrn():

    feat = 'segment_radii'

    assert_allclose(_stats(get_feature(feat, NRN)),
                    (0.12087134271860123,
                        1.0343990325927734,
                        507.01994501426816,
                        0.60359517263603357))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.all)),
                    (0.12087134271860123,
                        1.0343990325927734,
                        507.01994501426816,
                        0.60359517263603357))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.apical_dendrite)),
                    (0.13142434507608414,
                        1.0343990325927734,
                        123.41135908663273,
                        0.58767313850777492))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.basal_dendrite)),
                    (0.14712842553853989,
                        1.0215770602226257,
                        256.71241207793355,
                        0.61122002875698467))


def test_segment_meander_angles_pop():

    feat = 'segment_meander_angles'

    assert_allclose(_stats(get_feature(feat, POP)),
                    (0.0, 3.141592653589793, 14640.757557645244, 2.3958038876853616))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.all)),
                    (0.0, 3.141592653589793, 14640.757557645244, 2.3958038876853616))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.apical_dendrite)),
                    (0.3261019628980804, 3.093926149921232, 461.9816869933165, 2.4443475502291876))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.basal_dendrite)),
                    (0.0, 3.1415926535897931, 2926.2411975307768, 2.4084289691611334))


def test_segment_meander_angles_nrn():

    feat = 'segment_meander_angles'
    assert_allclose(_stats(get_feature(feat, NRN)),
                    (0.3261019628980804, 3.1299616642751253, 1842.3517793757057, 2.436973253142468))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.all)),
                    (0.3261019628980804, 3.1299616642751253, 1842.3517793757057, 2.436973253142468))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.apical_dendrite)),
                    (0.3261019628980804, 3.093926149921232, 461.9816869933165, 2.4443475502291876)
)

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.basal_dendrite)),
                    (0.47318725279312024, 3.129961675751181, 926.33847274926438, 2.4506308802890593))


def test_neurite_volumes_nrn():

    feat = 'neurite_volumes'

    assert_allclose(_stats(get_feature(feat, NRN)),
                    (271.94122143951864, 281.24754646913954, 1104.9077698137021, 276.22694245342552))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.all)),
                    (271.94122143951864, 281.24754646913954, 1104.9077698137021, 276.22694245342552))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.axon)),
                    (276.73860261723024, 276.73860261723024, 276.73860261723024, 276.73860261723024))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.basal_dendrite)),
                    (274.98039928781355, 281.24754646913954, 556.22794575695309, 278.11397287847655))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.apical_dendrite)),
                    (271.94122143951864, 271.94122143951864, 271.94122143951864, 271.94122143951864))


def test_neurite_volumes_pop():

    feat = 'neurite_volumes'

    assert_allclose(_stats(get_feature(feat, POP)),
                    (28.356406629821159, 281.24754646913954, 2249.4613918388391, 224.9461391838839))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.all)),
                    (28.356406629821159, 281.24754646913954, 2249.4613918388391, 224.9461391838839))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.axon)),
                    (276.58135508666612, 277.5357232437392, 830.85568094763551, 276.95189364921185))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.basal_dendrite)),
                    (28.356406629821159, 281.24754646913954, 1146.6644894516851, 191.1107482419475))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.apical_dendrite)),
                    (271.94122143951864, 271.94122143951864, 271.94122143951864, 271.94122143951864))


def test_neurite_density_nrn():

    feat = 'neurite_volume_density'

    assert_allclose(_stats(get_feature(feat, NRN)),
                    (0.24068543213643726, 0.52464681266899216, 1.4657913638494682, 0.36644784096236704))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.all)),
                    (0.24068543213643726, 0.52464681266899216, 1.4657913638494682, 0.36644784096236704))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.axon)),
                    (0.26289304906104355, 0.26289304906104355, 0.26289304906104355, 0.26289304906104355))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.basal_dendrite)),
                    (0.24068543213643726, 0.52464681266899216, 0.76533224480542938, 0.38266612240271469))

    assert_allclose(_stats(get_feature(feat, NRN, neurite_type=NeuriteType.apical_dendrite)),
                    (0.43756606998299519, 0.43756606998299519, 0.43756606998299519, 0.43756606998299519))


def test_neurite_density_pop():

    feat = 'neurite_volume_density'

    assert_allclose(_stats(get_feature(feat, POP)),
                    (6.1847539631150784e-06, 0.52464681266899216, 1.9767794901940539, 0.19767794901940539))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.all)),
                    (6.1847539631150784e-06, 0.52464681266899216, 1.9767794901940539, 0.19767794901940539))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.axon)),
                    (6.1847539631150784e-06, 0.26465213325053372, 0.5275513670655404, 0.17585045568851346),
                    rtol=1e-6)

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.basal_dendrite)),
                    (0.00034968816544949771, 0.52464681266899216, 1.0116620531455183, 0.16861034219091972))

    assert_allclose(_stats(get_feature(feat, POP, neurite_type=NeuriteType.apical_dendrite)),
                    (0.43756606998299519, 0.43756606998299519, 0.43756606998299519, 0.43756606998299519))


def test_segment_meander_angles_single_section():
    feat = 'segment_meander_angles'

    nrn = nm.load_neuron(StringIO(u"""((CellBody) (0 0 0 0))
                                      ((Dendrite)
                                       (0 0 0 2)
                                       (1 0 0 2)
                                       (1 1 0 2)
                                       (2 1 0 2)
                                       (2 2 0 2)))"""), reader='asc')

    nrt = nrn.neurites[0]
    pop = core.Population([nrn])

    ref = [math.pi / 2, math.pi / 2, math.pi / 2]

    nt.eq_(ref, get_feature(feat, nrt).tolist())
    nt.eq_(ref, get_feature(feat, nrn).tolist())
    nt.eq_(ref, get_feature(feat, pop).tolist())


def test_neurite_features_accept_single_tree():

    features = NEURITEFEATURES.keys()

    for f in features:
        ret = get_feature(f, NRN.neurites[0])
        nt.ok_(ret.dtype.kind in ('i', 'f'))
        nt.ok_(len(ret) > 0)


def test_register_neurite_feature_nrns():

    def npts(neurite):
        return len(neurite.points)

    def vol(neurite):
        return neurite.volume

    features.register_neurite_feature('foo', npts)

    n_points_ref = [len(n.points) for n in iter_neurites(NRNS)]
    n_points = get_feature('foo', NRNS)
    assert_items_equal(n_points, n_points_ref)

    # test neurite type filtering
    n_points_ref = [len(n.points) for n in iter_neurites(NRNS, filt=_is_type(NeuriteType.axon))]
    n_points = get_feature('foo', NRNS, neurite_type=NeuriteType.axon)
    assert_items_equal(n_points, n_points_ref)

    features.register_neurite_feature('bar', vol)

    n_volume_ref = [n.volume for n in iter_neurites(NRNS)]
    n_volume = get_feature('bar', NRNS)
    assert_items_equal(n_volume, n_volume_ref)

    # test neurite type filtering
    n_volume_ref = [n.volume for n in iter_neurites(NRNS, filt=_is_type(NeuriteType.axon))]
    n_volume = get_feature('bar', NRNS, neurite_type=NeuriteType.axon)
    assert_items_equal(n_volume, n_volume_ref)


def test_register_neurite_feature_pop():

    def npts(neurite):
        return len(neurite.points)

    def vol(neurite):
        return neurite.volume

    features.register_neurite_feature('foo', npts)

    n_points_ref = [len(n.points) for n in iter_neurites(POP)]
    n_points = get_feature('foo', POP)
    assert_items_equal(n_points, n_points_ref)

    # test neurite type filtering
    n_points_ref = [len(n.points) for n in iter_neurites(POP,
                                                         filt=_is_type(NeuriteType.basal_dendrite))]
    n_points = get_feature('foo', POP, neurite_type=NeuriteType.basal_dendrite)
    assert_items_equal(n_points, n_points_ref)

    features.register_neurite_feature('bar', vol)

    n_volume_ref = [n.volume for n in iter_neurites(POP)]
    n_volume = get_feature('bar', POP)
    assert_items_equal(n_volume, n_volume_ref)

    # test neurite type filtering
    n_volume_ref = [n.volume for n in iter_neurites(POP, filt=_is_type(NeuriteType.basal_dendrite))]
    n_volume = get_feature('bar', POP, neurite_type=NeuriteType.basal_dendrite)
    assert_items_equal(n_volume, n_volume_ref)


@nt.raises(NeuroMError)
def test_register_existing_feature_raises():
    features.register_neurite_feature('total_length', lambda n: None)


def test_section_lengths():
    ref_seclen = [n.length for n in iter_sections(NEURON)]
    seclen = get_feature('section_lengths', NEURON)
    nt.eq_(len(seclen), 84)
    assert_allclose(seclen, ref_seclen)


def test_section_lengths_axon():
    s = get_feature('section_lengths', NEURON, neurite_type=NeuriteType.axon)
    nt.eq_(len(s), 21)


def test_total_lengths_basal():
    s = get_feature('section_lengths', NEURON, neurite_type=NeuriteType.basal_dendrite)
    nt.eq_(len(s), 42)


def test_section_lengths_apical():
    s = get_feature('section_lengths', NEURON, neurite_type=NeuriteType.apical_dendrite)
    nt.eq_(len(s), 21)


def test_total_length_per_neurite_axon():
    tl = get_feature('total_length_per_neurite', NEURON, neurite_type=NeuriteType.axon)
    nt.eq_(len(tl), 1)
    assert_allclose(tl, (207.87975221))


def test_total_length_per_neurite_basal():
    tl = get_feature('total_length_per_neurite', NEURON, neurite_type=NeuriteType.basal_dendrite)
    nt.eq_(len(tl), 2)
    assert_allclose(tl, (211.11737442, 207.31504202))


def test_total_length_per_neurite_apical():
    tl = get_feature('total_length_per_neurite', NEURON, neurite_type=NeuriteType.apical_dendrite)
    nt.eq_(len(tl), 1)
    assert_allclose(tl, (214.37304578))


def test_total_length_axon():
    tl = get_feature('total_length', NEURON, neurite_type=NeuriteType.axon)
    nt.eq_(len(tl), 1)
    assert_allclose(tl, (207.87975221))


def test_total_length_basal():
    tl = get_feature('total_length', NEURON, neurite_type=NeuriteType.basal_dendrite)
    nt.eq_(len(tl), 1)
    assert_allclose(tl, (418.43241644))


def test_total_length_apical():
    tl = get_feature('total_length', NEURON, neurite_type=NeuriteType.apical_dendrite)
    nt.eq_(len(tl), 1)
    assert_allclose(tl, (214.37304578))


def test_section_lengths_invalid():
    s = get_feature('section_lengths', NEURON, neurite_type=NeuriteType.soma)
    nt.eq_(len(s), 0)
    s = get_feature('section_lengths', NEURON, neurite_type=NeuriteType.undefined)
    nt.eq_(len(s), 0)


def test_section_path_distances_axon():
    path_lengths = get_feature('section_path_distances', NEURON, neurite_type=NeuriteType.axon)
    nt.eq_(len(path_lengths), 21)


def test_segment_lengths():
    ref_seglen = nf.segment_lengths(NEURON)
    seglen = get_feature('segment_lengths', NEURON)
    nt.eq_(len(seglen), 840)
    assert_allclose(seglen, ref_seglen)

    seglen = get_feature('segment_lengths', NEURON, neurite_type=NeuriteType.all)
    nt.eq_(len(seglen), 840)
    assert_allclose(seglen, ref_seglen)


def test_local_bifurcation_angles():
    ref_local_bifangles = list(nf.local_bifurcation_angles(NEURON))

    local_bifangles = get_feature('local_bifurcation_angles', NEURON)
    nt.eq_(len(local_bifangles), 40)
    assert_allclose(local_bifangles, ref_local_bifangles)
    local_bifangles = get_feature('local_bifurcation_angles', NEURON, neurite_type=NeuriteType.all)
    nt.eq_(len(local_bifangles), 40)
    assert_allclose(local_bifangles, ref_local_bifangles)

    s = get_feature('local_bifurcation_angles', NEURON, neurite_type=NeuriteType.axon)
    nt.eq_(len(s), 10)

    s = get_feature('local_bifurcation_angles', NEURON, neurite_type=NeuriteType.basal_dendrite)
    nt.eq_(len(s), 20)

    s = get_feature('local_bifurcation_angles', NEURON, neurite_type=NeuriteType.apical_dendrite)
    nt.eq_(len(s), 10)


def test_local_bifurcation_angles_invalid():
    s = get_feature('local_bifurcation_angles', NEURON, neurite_type=NeuriteType.soma)
    nt.eq_(len(s), 0)
    s = get_feature('local_bifurcation_angles', NEURON, neurite_type=NeuriteType.undefined)
    nt.eq_(len(s), 0)


def test_remote_bifurcation_angles():
    ref_remote_bifangles = list(nf.remote_bifurcation_angles(NEURON))
    remote_bifangles = get_feature('remote_bifurcation_angles', NEURON)
    nt.eq_(len(remote_bifangles), 40)
    assert_allclose(remote_bifangles, ref_remote_bifangles)
    remote_bifangles = get_feature('remote_bifurcation_angles',
                                   NEURON, neurite_type=NeuriteType.all)
    nt.eq_(len(remote_bifangles), 40)
    assert_allclose(remote_bifangles, ref_remote_bifangles)

    s = get_feature('remote_bifurcation_angles', NEURON, neurite_type=NeuriteType.axon)
    nt.eq_(len(s), 10)

    s = get_feature('remote_bifurcation_angles', NEURON, neurite_type=NeuriteType.basal_dendrite)
    nt.eq_(len(s), 20)

    s = get_feature('remote_bifurcation_angles', NEURON, neurite_type=NeuriteType.apical_dendrite)
    nt.eq_(len(s), 10)


def test_remote_bifurcation_angles_invalid():
    s = get_feature('remote_bifurcation_angles', NEURON, neurite_type=NeuriteType.soma)
    nt.eq_(len(s), 0)
    s = get_feature('remote_bifurcation_angles', NEURON, neurite_type=NeuriteType.undefined)
    nt.eq_(len(s), 0)


def test_segment_radial_distances_origin():
    origin = (-100, -200, -300)
    ref_segs = nf.segment_radial_distances(NEURON)
    ref_segs_origin = nf.segment_radial_distances(NEURON, origin=origin)

    rad_dists = get_feature('segment_radial_distances', NEURON)
    rad_dists_origin = get_feature('segment_radial_distances', NEURON, origin=origin)

    nt.ok_(np.all(rad_dists == ref_segs))
    nt.ok_(np.all(rad_dists_origin == ref_segs_origin))
    nt.ok_(np.all(rad_dists_origin != ref_segs))

    nrns = [nm.load_neuron(Path(SWC_PATH, f)) for
            f in ('point_soma_single_neurite.swc', 'point_soma_single_neurite2.swc')]
    pop = Population(nrns)
    rad_dist_nrns = []
    for nrn in nrns:
        rad_dist_nrns.extend(nm.get('segment_radial_distances', nrn))

    rad_dist_nrns = np.array(rad_dist_nrns)
    rad_dist_pop = nm.get('segment_radial_distances', pop)
    assert_allclose(rad_dist_nrns, rad_dist_pop)


def test_section_radial_distances_endpoint():
    ref_sec_rad_dist = nf.section_radial_distances(NEURON)

    rad_dists = get_feature('section_radial_distances', NEURON)

    nt.eq_(len(rad_dists), 84)
    nt.ok_(np.all(rad_dists == ref_sec_rad_dist))

    nrns = [nm.load_neuron(Path(SWC_PATH, f)) for
            f in ('point_soma_single_neurite.swc', 'point_soma_single_neurite2.swc')]
    pop = Population(nrns)
    rad_dist_nrns = [nm.get('section_radial_distances', nrn) for nrn in nrns]
    rad_dist_pop = nm.get('section_radial_distances', pop)
    assert_items_equal(rad_dist_pop, rad_dist_nrns)


def test_section_radial_distances_origin():
    origin = (-100, -200, -300)
    ref_sec_rad_dist_origin = nf.section_radial_distances(NEURON, origin=origin)
    rad_dists = get_feature('section_radial_distances', NEURON, origin=origin)
    nt.eq_(len(rad_dists), 84)
    nt.ok_(np.all(rad_dists == ref_sec_rad_dist_origin))


def test_section_radial_axon():
    rad_dists = get_feature('section_radial_distances', NEURON, neurite_type=NeuriteType.axon)
    nt.eq_(len(rad_dists), 21)


def test_number_of_sections_all():
    assert_allclose(get_feature('number_of_sections', NEURON)[0], 84)
    assert_allclose(get_feature('number_of_sections', NEURON, neurite_type=NeuriteType.all)[0], 84)


def test_number_of_sections_axon():
    assert_allclose(get_feature('number_of_sections', NEURON, neurite_type=NeuriteType.axon)[0], 21)


def test_number_of_sections_basal():
    assert_allclose(get_feature('number_of_sections', NEURON, neurite_type=NeuriteType.basal_dendrite)[0], 42)


def test_n_sections_apical():
    assert_allclose(get_feature('number_of_sections', NEURON, neurite_type=NeuriteType.apical_dendrite)[0], 21)


def test_section_number_invalid():
    assert_allclose(get_feature('number_of_sections', NEURON, neurite_type=NeuriteType.soma)[0], 0)
    assert_allclose(get_feature('number_of_sections', NEURON, neurite_type=NeuriteType.undefined)[0], 0)


def test_per_neurite_number_of_sections():
    nsecs = get_feature('number_of_sections_per_neurite', NEURON)
    nt.eq_(len(nsecs), 4)
    nt.ok_(np.all(nsecs == [21, 21, 21, 21]))


def test_per_neurite_number_of_sections_axon():
    nsecs = get_feature('number_of_sections_per_neurite', NEURON, neurite_type=NeuriteType.axon)
    nt.eq_(len(nsecs), 1)
    nt.eq_(nsecs, [21])


def test_n_sections_per_neurite_basal():
    nsecs = get_feature('number_of_sections_per_neurite', NEURON,
                        neurite_type=NeuriteType.basal_dendrite)
    nt.eq_(len(nsecs), 2)
    nt.ok_(np.all(nsecs == [21, 21]))


def test_n_sections_per_neurite_apical():
    nsecs = get_feature('number_of_sections_per_neurite', NEURON,
                        neurite_type=NeuriteType.apical_dendrite)
    nt.eq_(len(nsecs), 1)
    nt.ok_(np.all(nsecs == [21]))


def test_neurite_number():
    assert_allclose(get_feature('number_of_neurites', NEURON)[0], 4)
    assert_allclose(get_feature('number_of_neurites', NEURON, neurite_type=NeuriteType.all)[0], 4)
    assert_allclose(get_feature('number_of_neurites', NEURON, neurite_type=NeuriteType.axon)[0], 1)
    assert_allclose(get_feature('number_of_neurites', NEURON, neurite_type=NeuriteType.basal_dendrite)[0], 2)
    assert_allclose(get_feature('number_of_neurites', NEURON, neurite_type=NeuriteType.apical_dendrite)[0], 1)
    assert_allclose(get_feature('number_of_neurites', NEURON, neurite_type=NeuriteType.soma)[0], 0)
    assert_allclose(get_feature('number_of_neurites', NEURON, neurite_type=NeuriteType.undefined)[0], 0)


def test_trunk_origin_radii():
    assert_allclose(get_feature('trunk_origin_radii', NEURON),
                    [0.85351288499400002,
                     0.18391483031299999,
                     0.66943255462899998,
                     0.14656092843999999])

    assert_allclose(get_feature('trunk_origin_radii', NEURON, neurite_type=NeuriteType.apical_dendrite),
                    [0.14656092843999999])
    assert_allclose(get_feature('trunk_origin_radii', NEURON, neurite_type=NeuriteType.basal_dendrite),
                    [0.18391483031299999,
                     0.66943255462899998])
    assert_allclose(get_feature('trunk_origin_radii', NEURON, neurite_type=NeuriteType.axon),
                    [0.85351288499400002])


def test_get_trunk_section_lengths():
    assert_allclose(get_feature('trunk_section_lengths', NEURON),
                    [9.579117366740002,
                     7.972322416776259,
                     8.2245287740603779,
                     9.212707985134525])
    assert_allclose(get_feature('trunk_section_lengths', NEURON, neurite_type=NeuriteType.apical_dendrite),
                    [9.212707985134525])
    assert_allclose(get_feature('trunk_section_lengths', NEURON, neurite_type=NeuriteType.basal_dendrite),
                    [7.972322416776259, 8.2245287740603779])
    assert_allclose(get_feature('trunk_section_lengths', NEURON, neurite_type=NeuriteType.axon),
                    [9.579117366740002])


def test_soma_radii():
    assert_allclose(get_feature('soma_radii', NEURON)[0], 0.13065629648763766)


def test_soma_surface_areas():
    area = 4. * math.pi * get_feature('soma_radii', NEURON)[0] ** 2
    assert_allclose(get_feature('soma_surface_areas', NEURON), area)


def test_sholl_frequency():
    assert_allclose(get_feature('sholl_frequency', NEURON),
                    [4, 8, 8, 14, 9, 8, 7, 7])

    assert_allclose(get_feature('sholl_frequency', NEURON, neurite_type=NeuriteType.all),
                    [4, 8, 8, 14, 9, 8, 7, 7])

    assert_allclose(get_feature('sholl_frequency', NEURON, neurite_type=NeuriteType.apical_dendrite),
                    [1, 2, 2, 2, 2, 2, 1, 1])

    assert_allclose(get_feature('sholl_frequency', NEURON, neurite_type=NeuriteType.basal_dendrite),
                    [2, 4, 4, 6, 5, 4, 4, 4])

    assert_allclose(get_feature('sholl_frequency', NEURON, neurite_type=NeuriteType.axon),
                    [1, 2, 2, 6, 2, 2, 2, 2])


@nt.nottest  # test_get_segment_lengths is disabled in test_get_features
def test_section_path_distances_endpoint():

    ref_sec_path_len_start = list(iter_neurites(NEURON, sec.start_point_path_length))
    ref_sec_path_len = list(iter_neurites(NEURON, sec.end_point_path_length))
    path_lengths = get_feature('section_path_distances', NEURON)
    nt.ok_(ref_sec_path_len != ref_sec_path_len_start)
    nt.eq_(len(path_lengths), 84)
    nt.ok_(np.all(path_lengths == ref_sec_path_len))


@nt.nottest  # test_get_segment_lengths is disabled in test_get_features
def test_section_path_distances_start_point():

    ref_sec_path_len_start = list(iter_neurites(NEURON, sec.start_point_path_length))
    path_lengths = get_feature('section_path_distances', NEURON, use_start_point=True)
    nt.eq_(len(path_lengths), 84)
    nt.ok_(np.all(path_lengths == ref_sec_path_len_start))


def test_partition():
    nt.ok_(np.all(get_feature('partition', NRNS)[:10] == np.array(
        [19.,  17.,  15.,  13.,  11.,   9.,   7.,   5.,   3.,   1.])))


def test_partition_asymmetry():
    assert_allclose(get_feature('partition_asymmetry', NRNS)[:10], np.array([0.9, 0.88888889, 0.875,
                                                                             0.85714286, 0.83333333,

                                                                             0.8, 0.75,  0.66666667,
                                                                             0.5,  0.]))


def test_section_strahler_orders():
    path = Path(SWC_PATH, 'strahler.swc')
    n = nm.load_neuron(path)
    assert_allclose(get_feature('section_strahler_orders', n),
                    [4, 1, 4, 3, 2, 1, 1, 2, 1, 1, 3, 1, 3, 2, 1, 1, 2, 1, 1])



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

""" Distribution configuration for neurom
"""
# pylint: disable=R0801
from setuptools import setup
from setuptools import find_packages


setup(
    description='NeuroM: a light-weight neuron morphology analysis package',
    author='Blue Brain Project, EPFL',
    url='http://https://github.com/BlueBrain/NeuroM',
    install_requires=[
        'click>=7.0',
        'h5py>=2.7.1',
        'matplotlib>=3.2.1',
        'morphio>=2.3.10',
        'numpy>=1.8.0',
        'pandas>=1.0.5',
        'pyyaml>=3.10',
        'scipy>=1.2.0',
        'tqdm>=4.8.4',
        ],
    packages=find_packages(),
    license='BSD',
    scripts=['apps/morph_check',
             'apps/morph_stats',
             ],
    entry_points={
        'console_scripts': ['neurom=neurom.apps.cli:cli']
    },
    name='neurom',
    extras_require={
        'plotly': ['plotly>=3.6.0'],
    },
    include_package_data=True,
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 6 - Mature',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    setup_requires=['setuptools_scm'],

    # 2.0.1 allows for a pip requirement formatted like "neurom>=2.0.0"
    version="2.0.1.dev4",
)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Neuron Population Classes and Functions."""

from itertools import chain


class Population(object):
    """Neuron Population Class.

    Features:
        - flattened collection of neurites.
        - collection of somas, neurons.
        - iterable-like iteration over neurons.
    """
    def __init__(self, neurons, name='Population'):
        """Construct a neuron population.

        Arguments:
            neurons: iterable of neuron objects.
            name: Optional name for this Population.
        """
        self.neurons = tuple(neurons)
        self.somata = tuple(neu.soma for neu in neurons)
        self.neurites = tuple(chain.from_iterable(neu.neurites for neu in neurons))
        self.name = name

    def __iter__(self):
        """Iterator to populations's neurons."""
        return iter(self.neurons)

    def __len__(self):
        """Length of neuron collection."""
        return len(self.neurons)

    def __getitem__(self, idx):
        """Get neuron at index idx."""
        return self.neurons[idx]

    def __str__(self):
        """Return a string representation."""
        return 'Population <name: %s, nneurons: %d>' % (self.name, len(self.neurons))



# Copyright (c) 2020, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Bifurcation point functions."""

import numpy as np
from neurom import morphmath
from neurom.exceptions import NeuroMError
from neurom.core.dataformat import COLS
from neurom.features import sectionfunc


def _raise_if_not_bifurcation(section):
    n_children = len(section.children)
    if n_children != 2:
        raise NeuroMError('A bifurcation point must have exactly 2 children, found {}'.format(
            n_children))


def local_bifurcation_angle(bif_point):
    """Return the opening angle between two out-going sections in a bifurcation point.

    We first ensure that the input point has only two children.

    The bifurcation angle is defined as the angle between the first non-zero
    length segments of a bifurcation point.
    """
    def skip_0_length(sec):
        """Return the first point with non-zero distance to first point."""
        p0 = sec[0]
        cur = sec[1]
        for i, p in enumerate(sec[1:]):
            if not np.all(p[:COLS.R] == p0[:COLS.R]):
                cur = sec[i + 1]
                break

        return cur

    _raise_if_not_bifurcation(bif_point)

    ch0, ch1 = (skip_0_length(bif_point.children[0].points),
                skip_0_length(bif_point.children[1].points))

    return morphmath.angle_3points(bif_point.points[-1], ch0, ch1)


def remote_bifurcation_angle(bif_point):
    """Return the opening angle between two out-going sections in a bifurcation point.

    We first ensure that the input point has only two children.

    The angle is defined as between the bifurcation point and the
    last points in the out-going sections.
    """
    _raise_if_not_bifurcation(bif_point)

    return morphmath.angle_3points(bif_point.points[-1],
                                   bif_point.children[0].points[-1],
                                   bif_point.children[1].points[-1])


def bifurcation_partition(bif_point):
    """Calculate the partition at a bifurcation point.

    We first ensure that the input point has only two children.

    The number of nodes in each child tree is counted. The partition is
    defined as the ratio of the largest number to the smallest number.
    """
    _raise_if_not_bifurcation(bif_point)

    n = float(sum(1 for _ in bif_point.children[0].ipreorder()))
    m = float(sum(1 for _ in bif_point.children[1].ipreorder()))
    return max(n, m) / min(n, m)


def partition_asymmetry(bif_point):
    """Calculate the partition asymmetry at a bifurcation point.

    Partition asymmetry is defined in https://www.ncbi.nlm.nih.gov/pubmed/18568015

    The number of nodes in each child tree is counted. The partition
    is defined as the ratio of the absolute difference and the sum
    of the number of bifurcations in the two child subtrees
    at each branch point.
    """
    _raise_if_not_bifurcation(bif_point)

    n = float(sum(1 for _ in bif_point.children[0].ipreorder()))
    m = float(sum(1 for _ in bif_point.children[1].ipreorder()))
    if n == m:
        return 0.0
    return abs(n - m) / abs(n + m)


def partition_pair(bif_point):
    """Calculate the partition pairs at a bifurcation point.

    The number of nodes in each child tree is counted. The partition
    pairs is the number of bifurcations in the two child subtrees
    at each branch point.
    """
    n = float(sum(1 for _ in bif_point.children[0].ipreorder()))
    m = float(sum(1 for _ in bif_point.children[1].ipreorder()))
    return (n, m)


def sibling_ratio(bif_point, method='first'):
    """Calculate the sibling ratio of a bifurcation point.

    The sibling ratio is the ratio between the diameters of the
    smallest and the largest child. It is a real number between
    0 and 1. Method argument allows one to consider mean diameters
    along the child section instead of diameter of the first point.
    """
    _raise_if_not_bifurcation(bif_point)

    if method not in {'first', 'mean'}:
        raise ValueError('Please provide a valid method for sibling ratio, found %s' % method)

    if method == 'first':
        n = bif_point.children[0].points[0, COLS.R]
        m = bif_point.children[1].points[0, COLS.R]
    if method == 'mean':
        n = sectionfunc.section_mean_radius(bif_point.children[0])
        m = sectionfunc.section_mean_radius(bif_point.children[1])
    return min(n, m) / max(n, m)


def diameter_power_relation(bif_point, method='first'):
    """Calculate the diameter power relation at a bifurcation point.

    The diameter power relation is defined in https://www.ncbi.nlm.nih.gov/pubmed/18568015

    This quantity gives an indication of how far the branching is from
    the Rall ratio

    diameter_power_relation==1 means perfect Rall ratio
    """
    _raise_if_not_bifurcation(bif_point)

    if method not in {'first', 'mean'}:
        raise ValueError('Please provide a valid method for sibling ratio, found %s' % method)

    if method == 'first':
        d_child = bif_point.points[-1, COLS.R]
        d_child1 = bif_point.children[0].points[0, COLS.R]
        d_child2 = bif_point.children[1].points[0, COLS.R]
    if method == 'mean':
        d_child = sectionfunc.section_mean_radius(bif_point)
        d_child1 = sectionfunc.section_mean_radius(bif_point.children[0])
        d_child2 = sectionfunc.section_mean_radius(bif_point.children[1])
    return (d_child / d_child1)**(1.5) + (d_child / d_child2)**(1.5)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pathlib import Path

import numpy as np

from neurom import load_neuron, NeuriteType
from neurom.core.dataformat import COLS
from numpy.testing import assert_array_equal, assert_equal

from nose import tools as nt
from neurom.exceptions import SomaError, RawDataError


DATA_PATH = Path(__file__).parent.parent.parent.parent / 'test_data'
SWC_PATH = Path(DATA_PATH, 'swc')
SWC_SOMA_PATH = Path(SWC_PATH, 'soma')


@nt.raises(RawDataError)
def test_repeated_id():
    n = load_neuron(SWC_PATH / 'repeated_id.swc')


@nt.raises(SomaError)
def test_neurite_followed_by_soma():
    load_neuron(SWC_PATH / 'soma_with_neurite_parent.swc')


def test_read_single_neurite():
    n = load_neuron(SWC_PATH / 'point_soma_single_neurite.swc')
    nt.eq_(len(n.neurites), 1)
    nt.eq_(n.neurites[0].root_node.id, 0)
    assert_array_equal(n.soma.points,
                       [[0, 0, 0, 3.0]])
    nt.eq_(len(n.neurites), 1)
    nt.eq_(len(n.sections), 1)
    assert_array_equal(n.neurites[0].points,
                       np.array([[0, 0, 2, 0.5],
                                 [0, 0, 3, 0.5],
                                 [0, 0, 4, 0.5],
                                 [0, 0, 5, 0.5]]))


def test_read_split_soma():
    n = load_neuron(SWC_PATH / 'split_soma_two_neurites.swc')

    assert_array_equal(n.soma.points,
                       [[1, 0, 1, 4.0],
                        [2, 0, 0, 4.0],
                        [3, 0, 0, 4.0]])

    nt.assert_equal(len(n.neurites), 2)
    assert_array_equal(n.neurites[0].points,
                       [[0, 0, 2, 0.5],
                        [0, 0, 3, 0.5],
                        [0, 0, 4, 0.5],
                        [0, 0, 5, 0.5]])

    assert_array_equal(n.neurites[1].points,
                       [[0, 0, 6, 0.5],
                        [0, 0, 7, 0.5],
                        [0, 0, 8, 0.5],
                        [0, 0, 9, 0.5]])

    nt.eq_(len(n.sections), 2)


def test_weird_indent():

    n = load_neuron("""

                 # this is the same as simple.swc

# but with a questionable styling

     1 1  0  0 0 1. -1
 2 3  0  0 0 1.  1

 3 3  0  5 0 1.  2
 4 3 -5  5 0 0.  3



 5 3  6  5 0 0.  3
     6 2  0  0 0 1.  1
 7 2  0 -4 0 1.  6

 8 2  6 -4 0         0.  7
 9 2 -5 -4 0 0.  7
""", reader='swc')

    simple = load_neuron(SWC_PATH / 'simple.swc')
    assert_array_equal(simple.points,
                       n.points)


@nt.raises(RawDataError)
def test_cyclic():
    load_neuron("""
    1 1  0  0 0 1. -1
    2 3  0  0 0 1.  1
    3 3  0  5 0 1.  2
    4 3 -5  5 0 0.  3
    5 3  6  5 0 0.  3
    6 2  0  0 0 1.  6  # <-- cyclic point
    7 2  0 -4 0 1.  6
    8 2  6 -4 0 0.  7
    9 2 -5 -4 0 0.  7""", reader='swc')


def test_simple_reversed():
    n = load_neuron(SWC_PATH / 'simple_reversed.swc')
    assert_array_equal(n.soma.points,
                       [[0, 0, 0, 1]])
    nt.assert_equal(len(n.neurites), 2)
    nt.assert_equal(len(n.neurites[0].points), 4)
    assert_array_equal(n.neurites[0].points,
                       [[0, 0, 0, 1],
                        [0, 5, 0, 1],
                        [-5, 5, 0, 0],
                        [6, 5, 0, 0]])
    assert_array_equal(n.neurites[1].points,
                       [[0, 0, 0, 1],
                        [0, -4, 0, 1],
                        [6, -4, 0, 0],
                        [-5, -4, 0, 0]])

# def test_custom_type():
#     neuron = load_neuron(Path(SWC_PATH, 'custom_type.swc'))
#     assert_equal(neuron.neurites[1].type, NeuriteType.custom)

def test_undefined_type():
    neuron = load_neuron(Path(SWC_PATH, 'undefined_type.swc'))
    assert_equal(neuron.neurites[1].type, NeuriteType.undefined)



#!/usr/bin/env python
# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Box Plot function for multiple neurons."""

from neurom.view import common


def boxplot(neurons, feature, new_fig=True, subplot=False):
    """Plot a histogram of the selected feature for the population of neurons.
    Plots x-axis versus y-axis on a scatter|histogram|binned values plot.

    More information about the plot and how it works.

    Parameters
    ----------
    neurons : list
        List of Neurons. Single neurons must be encapsulated in a list.

    feature : str
    The feature of interest.

    Options
    -------

    subplot : bool
        Default is False, which returns a matplotlib figure object. If True,
        returns a matplotlib axis object, for use as a subplot.
    """
    feature_values = [getattr(neu, 'get_' + feature)() for neu in neurons]

    _, ax = common.get_figure(new_fig=new_fig, subplot=subplot)

    ax.boxplot(feature_values)

    x_labels = ['neuron_id' for _ in neurons]

    ax.set_xticklabels(x_labels)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
from copy import copy
from pathlib import Path

from nose import tools as nt

from neurom.check.runner import CheckRunner
from neurom.exceptions import ConfigError

from collections import OrderedDict
_path = os.path.dirname(os.path.abspath(__file__))
SWC_PATH = Path(__file__).parent / '../../../test_data/swc/'


CONFIG = {
    'checks': {
        'structural_checks': [
            #'is_single_tree',
            'has_soma_points',
            # 'has_sequential_ids',
            # 'has_increasing_ids',
            'has_valid_soma',
            # 'has_valid_neurites'
        ],
        'neuron_checks': [
            'has_basal_dendrite',
            'has_axon',
            'has_apical_dendrite',
            'has_all_nonzero_segment_lengths',
            'has_all_nonzero_section_lengths',
            'has_all_nonzero_neurite_radii',
            'has_nonzero_soma_radius'
        ]
    },
    'options': {
        'has_nonzero_soma_radius': 0.0,
        "has_all_nonzero_neurite_radii": 0.007,
        "has_all_nonzero_segment_lengths": 0.01,
        "has_all_nonzero_section_lengths": [0.01]
    },
}

CONFIG_COLOR = copy(CONFIG)
CONFIG_COLOR['color'] = True



def _run_test(path, ref, config=CONFIG, should_pass=False):
    '''Run checkers with the passed "config" on file "path"
    and compare the results to "ref"'''
    results = CheckRunner(config).run(path)
    for item, item_ref in zip(results['files'][path].items(), ref.items()):
        nt.assert_equal(item, item_ref)
    nt.assert_equal(results['STATUS'],
                    "PASS" if should_pass else "FAIL")

ref = OrderedDict([
    # "Is single tree": True,
    ("Has soma points", True),
    # "Has sequential ids": True,
    # "Has increasing ids": True,
    ("Has valid soma", True),
    # "Has valid neurites": True,
    ("Has basal dendrite", True),
    ("Has axon", True),
    ("Has apical dendrite", True),
    ("Has all nonzero segment lengths", True),
    ("Has all nonzero section lengths", True),
    ("Has all nonzero neurite radii", True),
    ("Has nonzero soma radius", True),
    ("ALL", True)
])

def test_ok_neuron():
    _run_test(os.path.join(SWC_PATH, 'Neuron.swc'),
              ref,
              should_pass=True)

def test_ok_neuron_color():
    _run_test(os.path.join(SWC_PATH, 'Neuron.swc'),
              ref,
              CONFIG_COLOR,
              should_pass=True)


def test_zero_length_sections_neuron():
    _run_test(os.path.join(SWC_PATH, 'Neuron_zero_length_sections.swc'),
              OrderedDict([
                  # "Is single tree": True,
                  ("Has soma points", True),
                  # "Has sequential ids": True,
                  # "Has increasing ids": True,
                  ("Has valid soma", True),
                  # "Has valid neurites": True,
                  ("Has basal dendrite", True),
                  ("Has axon", True),
                  ("Has apical dendrite", True),
                  ("Has all nonzero segment lengths", False),
                  ("Has all nonzero section lengths", False),
                  ("Has all nonzero neurite radii", True),
                  ("Has nonzero soma radius", True),
                  ("ALL", False)
              ]))


def test_single_apical_neuron():
    _run_test(os.path.join(SWC_PATH, 'Single_apical.swc'),
              OrderedDict([
                  # "Is single tree": True,
                  ("Has soma points", True),
                  # "Has sequential ids": True,
                  # "Has increasing ids": True,
                  ("Has valid soma", True),
                  # "Has valid neurites": True,
                  ("Has basal dendrite", False),
                  ("Has axon", False),
                  ("Has apical dendrite", True),
                  ("Has all nonzero segment lengths", False),
                  ("Has all nonzero section lengths", True),
                  ("Has all nonzero neurite radii", True),
                  ("Has nonzero soma radius", True),
                  ("ALL", False)
              ]))


def test_single_basal_neuron():
    _run_test(os.path.join(SWC_PATH, 'Single_basal.swc'),
              OrderedDict(
                  ([
                      # "Is single tree": True,
                      ("Has soma points", True),
                      # "Has sequential ids": True,
                      # "Has increasing ids": True,
                      ("Has valid soma", True),
                      # "Has valid neurites": True,
                      ("Has basal dendrite", True),
                      ("Has axon", False),
                      ("Has apical dendrite", False),
                      ("Has all nonzero segment lengths", False),
                      ("Has all nonzero section lengths", True),
                      ("Has all nonzero neurite radii", True),
                      ("Has nonzero soma radius", True),
                      ("ALL", False)
                  ])))


def test_single_axon_neuron():
    _run_test(os.path.join(SWC_PATH, 'Single_axon.swc'),
              OrderedDict([
                  # "Is single tree": True,
                  ("Has soma points", True),
                  # "Has sequential ids": True,
                  # "Has increasing ids": True,
                  ("Has valid soma", True),
                  # "Has valid neurites": True,
                  ("Has basal dendrite", False),
                  ("Has axon", True),
                  ("Has apical dendrite", False),
                  ("Has all nonzero segment lengths", False),
                  ("Has all nonzero section lengths", True),
                  ("Has all nonzero neurite radii", True),
                  ("Has nonzero soma radius", True),
                  ("ALL", False)
              ]))


def test_single_apical_no_soma():
    import neurom
    n=neurom.load_neuron(os.path.join(SWC_PATH, 'Single_apical_no_soma.swc'))
    _run_test(os.path.join(SWC_PATH, 'Single_apical_no_soma.swc'),
              OrderedDict([
                  # "Is single tree": True,
                  ("Has soma points", False),
                  # "Has sequential ids": True,
                  # "Has increasing ids": True,
                  ("Has valid soma", False),
                  # "Has valid neurites": False,
                  ("Has basal dendrite", False),
                  ("Has axon", False),
                  ("Has apical dendrite", True),
                  ("Has all nonzero segment lengths", False),
                  ("Has all nonzero section lengths", True),
                  ("Has all nonzero neurite radii", True),
                  ("Has nonzero soma radius", False),
                  ("ALL", False)
              ]))


def test_directory_input():
    checker = CheckRunner(CONFIG)
    summ = checker.run(SWC_PATH)
    nt.eq_(summ['files'][os.path.join(SWC_PATH, 'Single_axon.swc')]['Has axon'], True)
    nt.eq_(summ['files'][os.path.join(SWC_PATH, 'Single_apical_no_soma.swc')]['Has axon'], False)


@nt.raises(IOError)
def test_invalid_data_path_raises_IOError():
    checker = CheckRunner(CONFIG)
    _ = checker.run('foo/bar/baz')


def test__sanitize_config():
    # fails if missing 'checks'
    nt.assert_raises(ConfigError, CheckRunner._sanitize_config, {})

    # creates minimal config
    new_config = CheckRunner._sanitize_config({'checks': {}})
    nt.eq_(new_config, {'checks':
                        {'structural_checks': [],
                         'neuron_checks': [],
                         },
                        'options': {},
                        'color': False,
                        })

    # makes no changes to already filled out config
    new_config = CheckRunner._sanitize_config(CONFIG)
    nt.eq_(CONFIG, new_config)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import itertools as it
from pathlib import Path
import warnings

import numpy as np
from nose.tools import assert_raises, eq_, ok_
from numpy.testing import assert_allclose, assert_array_almost_equal
from io import StringIO

from neurom import load_neuron
from neurom.view import view, common
from neurom.core import Section
from neurom.core._soma import make_soma
from neurom.core.types import NeuriteType
from morphio import SomaType

from .utils import (  # needs to be at top to trigger matplotlib Agg backend
    get_fig_2d, get_fig_3d)

DATA_PATH = './test_data'
SWC_PATH = Path(DATA_PATH, 'swc/')
fst_neuron = load_neuron(Path(SWC_PATH, 'Neuron.swc'))
simple_neuron = load_neuron(Path(SWC_PATH, 'simple.swc'))


def test_tree():
    neuron = load_neuron(Path(SWC_PATH, 'simple-different-section-types.swc'))
    expected_colors = {'black': np.array([[0., 0., 0., 1.] for _ in range(3)]),
                       None: [[1.      , 0.      , 0.      , 1.],
                              [1.      , 0.      , 0.      , 1.],
                              [0.501961, 0.      , 0.501961, 1.]]}
    for input_color, expected_colors in expected_colors.items():
        with get_fig_2d() as (fig, ax):
            tree = neuron.neurites[0]
            view.plot_tree(ax, tree,
                           color=input_color, diameter_scale=None, alpha=1., linewidth=1.2)
            collection = ax.collections[0]
            eq_(collection.get_linewidth()[0], 1.2)
            assert_array_almost_equal(collection.get_colors(), expected_colors)

        with get_fig_2d() as (fig, ax):
            tree = neuron.neurites[0]
            view.plot_tree(ax, tree,
                           color=input_color, alpha=1., linewidth=1.2, realistic_diameters=True)
            collection = ax.collections[0]
            eq_(collection.get_linewidth()[0], 1.0)
            assert_array_almost_equal(collection.get_facecolors(), expected_colors)


    with get_fig_2d() as (fig, ax):
        assert_raises(AssertionError, view.plot_tree, ax, tree, plane='wrong')

    with get_fig_2d() as (fig, ax):
        tree = simple_neuron.neurites[0]
        view.plot_tree(ax, tree)
        np.testing.assert_allclose(ax.dataLim.bounds, (-5., 0., 11., 5.), atol=5)


def test_neuron():
    with get_fig_2d() as (fig, ax):
        view.plot_neuron(ax, fst_neuron)
        ok_(ax.get_title() == fst_neuron.name)
        assert_allclose(ax.dataLim.get_points(),
                                   [[-40.32853516, -57.600172],
                                    [64.74726272, 48.51626225], ])

    with get_fig_2d() as (fig, ax):
        assert_raises(AssertionError, view.plot_tree, ax, fst_neuron, plane='wrong')


def test_tree3d():
    with get_fig_3d() as (fig, ax):
        tree = simple_neuron.neurites[0]
        view.plot_tree3d(ax, tree)
        xy_bounds = ax.xy_dataLim.bounds
        np.testing.assert_allclose(xy_bounds, (-5., 0., 11., 5.), atol=5)
        zz_bounds = ax.zz_dataLim.bounds
        assert_allclose(zz_bounds, (0., 0., 1., 1.))


def test_neuron3d():
    with get_fig_3d() as (fig, ax):
        view.plot_neuron3d(ax, fst_neuron)
        ok_(ax.get_title() == fst_neuron.name)
        assert_allclose(ax.xy_dataLim.get_points(),
                                   [[-40.32853516, -57.600172],
                                    [64.74726272, 48.51626225], ])
        assert_allclose(ax.zz_dataLim.get_points().T[0],
                                   (-00.09999862, 54.20408797))


def test_neuron_no_neurites():
    filename = Path(SWC_PATH, 'point_soma.swc')
    with get_fig_2d() as (fig, ax):
        view.plot_neuron(ax, load_neuron(filename))


def test_neuron3d_no_neurites():
    filename = Path(SWC_PATH, 'point_soma.swc')
    with get_fig_3d() as (fig, ax):
        view.plot_neuron3d(ax, load_neuron(filename))


def test_dendrogram():
    with get_fig_2d() as (fig, ax):
        view.plot_dendrogram(ax, fst_neuron)
        assert_allclose(ax.get_xlim(), (-10., 180.), rtol=0.25)

    with get_fig_2d() as (fig, ax):
        view.plot_dendrogram(ax, fst_neuron, show_diameters=False)
        assert_allclose(ax.get_xlim(), (-10., 180.), rtol=0.25)


with warnings.catch_warnings(record=True):
    soma0 = fst_neuron.soma

    # upright, varying radius
    soma_2pt_normal = load_neuron(StringIO(u"""1 1 0  0 0 1  -1
                                               2 1 0 10 0 10  1"""), reader='swc').soma

    # upright, uniform radius, multiple cylinders
    soma_3pt_normal = load_neuron(StringIO(u"""1 1 0 -10 0 10  -1
                                               2 1 0   0 0 10   1
                                               3 1 0  10 0 10   2"""), reader='swc').soma

    # increasing radius, multiple cylinders
    soma_4pt_normal_cylinder = load_neuron(StringIO(u"""1 1   0   0   0 1 -1
                                                       2 1   0 -10   0 2  1
                                                       3 1   0 -10  10 4  2
                                                       4 1 -10 -10 -10 4  3"""), reader='swc').soma

    soma_4pt_normal_contour = load_neuron(StringIO(u"""((CellBody)
                                                       (0     0   0 1)
                                                       (0   -10   0 2)
                                                       (0   -10  10 4)
                                                       (-10 -10 -10 4))"""), reader='asc').soma


def test_soma():
    for s in (soma0,
              soma_3pt_normal,
              soma_4pt_normal_cylinder,
              soma_4pt_normal_contour):
        with get_fig_2d() as (fig, ax):
            view.plot_soma(ax, s)
            common.plt.close(fig)

        with get_fig_2d() as (fig, ax):
            view.plot_soma(ax, s, soma_outline=False)
            common.plt.close(fig)


def test_soma3d():
    with get_fig_3d() as (_, ax):
        view.plot_soma3d(ax, soma_3pt_normal)
        assert_allclose(ax.get_xlim(), (-11.,  11.), atol=2)
        assert_allclose(ax.get_ylim(), (-11.,  11.), atol=2)
        assert_allclose(ax.get_zlim(), (-10.,  10.), atol=2)


def test_get_color():
    eq_(view._get_color(None, NeuriteType.basal_dendrite), "red")
    eq_(view._get_color(None, NeuriteType.axon), "blue")
    eq_(view._get_color(None, NeuriteType.apical_dendrite), "purple")
    eq_(view._get_color(None, NeuriteType.soma), "black")
    eq_(view._get_color(None, NeuriteType.undefined), "green")
    eq_(view._get_color(None, 'wrong'), "green")
    eq_(view._get_color('blue', 'wrong'), "blue")
    eq_(view._get_color('yellow', NeuriteType.axon), "yellow")



from pathlib import Path
from collections import namedtuple

import numpy as np
from nose import tools as nt
from numpy.testing import assert_array_almost_equal

import neurom.view.dendrogram as dm
from neurom import load_neuron, get
from neurom.core.types import NeuriteType

DATA_PATH = Path(__file__).parent.parent.parent.parent / 'test_data'
NEURON_PATH = Path(DATA_PATH, 'h5', 'v1', 'Neuron.h5')


def test_create_dendrogram_neuron():
    neuron = load_neuron(NEURON_PATH)
    dendrogram = dm.Dendrogram(neuron)
    nt.assert_equal(NeuriteType.soma, dendrogram.neurite_type)
    soma_len = 1.0
    nt.assert_equal(soma_len, dendrogram.height)
    nt.assert_equal(soma_len, dendrogram.width)
    assert_array_almost_equal(
        [[-.5, 0], [-.5, soma_len], [.5, soma_len], [.5, 0]],
        dendrogram.coords)
    nt.assert_equal(len(neuron.neurites), len(dendrogram.children))


def test_dendrogram_get_coords():
    segment_lengts = np.array([0, 1, 1])
    segment_radii = np.array([.5, 1, .25])
    coords = dm.Dendrogram.get_coords(segment_lengts, segment_radii)
    assert_array_almost_equal(
        [[-.5, 0], [-1, 1], [-.25, 2], [.25, 2], [1, 1], [.5, 0]],
        coords)


def test_create_dendrogram_neurite():
    def assert_trees(neurom_section, dendrogram):
        nt.assert_equal(len(neurom_section.children), len(dendrogram.children))
        for i, d in enumerate(dendrogram.children):
            section = neurom_section.children[i]
            nt.assert_equal(section.type, d.neurite_type)

    neuron = load_neuron(NEURON_PATH)
    neurite = neuron.neurites[0]
    dendrogram = dm.Dendrogram(neurite)
    nt.assert_equal(neurite.type, dendrogram.neurite_type)
    assert_trees(neurite.root_node, dendrogram)


def test_move_positions():
    origin = [10, -10]
    positions = {1: [0, 0], 2: [3, -3]}
    moved_positions = dm.move_positions(positions, origin)
    nt.assert_list_equal(list(positions.keys()), list(moved_positions.keys()))
    nt.assert_list_equal([10, -10], moved_positions[1].tolist())
    nt.assert_list_equal([13, -13], moved_positions[2].tolist())


def test_get_size():
    DendrogramMock = namedtuple('Dendrogram', 'height')
    dendrogram1 = DendrogramMock(10)
    dendrogram2 = DendrogramMock(1)
    positions = {dendrogram1: [-1, 0], dendrogram2: [3, 3]}
    w, h = dm.get_size(positions)
    nt.assert_equal(4, w)
    nt.assert_equal(10, h)


def test_layout_dendrogram():
    def assert_layout(dendrogram):
        for i, child in enumerate(dendrogram.children):
            # child is higher than parent in Y coordinate
            nt.assert_greater_equal(
                positions[child][1],
                positions[dendrogram][1] + dendrogram.height)
            if i < len(dendrogram.children) - 1:
                next_child = dendrogram.children[i + 1]
                # X space between child is enough for their widths
                nt.assert_greater(
                    positions[next_child][0] - positions[child][0],
                    .5 * (next_child.width + child.width))
            assert_layout(child)

    neuron = load_neuron(NEURON_PATH)
    dendrogram = dm.Dendrogram(neuron)
    positions = dm.layout_dendrogram(dendrogram, np.array([0, 0]))
    assert_layout(dendrogram)


def test_neuron_not_corrupted():
    # Regression for #492: dendrogram was corrupting
    # neuron used to construct it.
    # This caused the section path distance calculation
    # to raise a KeyError exception.
    neuron = load_neuron(NEURON_PATH)
    dm.Dendrogram(neuron)
    nt.assert_greater(get('section_path_distances', neuron).size, 0)
